package com.softslate.commerce.businessobjects.payment;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.UUID;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.codehaus.jackson.map.ObjectMapper;
import org.codehaus.jackson.type.TypeReference;

import com.softslate.commerce.businessobjects.customer.Customer;
import com.softslate.commerce.businessobjects.customer.CustomerProcessor;
import com.softslate.commerce.customer.order.BillingAddressForm;
import com.softslate.commerce.daos.payment.PaymentDAO;

import net.authorize.Environment;
import net.authorize.api.contract.v1.ArrayOfSetting;
import net.authorize.api.contract.v1.CreateCustomerPaymentProfileRequest;
import net.authorize.api.contract.v1.CreateCustomerPaymentProfileResponse;
import net.authorize.api.contract.v1.CreateCustomerProfileRequest;
import net.authorize.api.contract.v1.CreateCustomerProfileResponse;
import net.authorize.api.contract.v1.CreateTransactionRequest;
import net.authorize.api.contract.v1.CreateTransactionResponse;
import net.authorize.api.contract.v1.CreditCardType;
import net.authorize.api.contract.v1.CustomerAddressType;
import net.authorize.api.contract.v1.CustomerDataType;
import net.authorize.api.contract.v1.CustomerPaymentProfileExType;
import net.authorize.api.contract.v1.CustomerPaymentProfileMaskedType;
import net.authorize.api.contract.v1.CustomerPaymentProfileType;
import net.authorize.api.contract.v1.CustomerProfilePaymentType;
import net.authorize.api.contract.v1.CustomerProfileType;
import net.authorize.api.contract.v1.CustomerTypeEnum;
import net.authorize.api.contract.v1.DeleteCustomerPaymentProfileRequest;
import net.authorize.api.contract.v1.DeleteCustomerPaymentProfileResponse;
import net.authorize.api.contract.v1.ExtendedAmountType;
import net.authorize.api.contract.v1.GetCustomerProfileRequest;
import net.authorize.api.contract.v1.GetCustomerProfileResponse;
import net.authorize.api.contract.v1.GetHostedPaymentPageRequest;
import net.authorize.api.contract.v1.GetHostedPaymentPageResponse;
import net.authorize.api.contract.v1.GetHostedProfilePageRequest;
import net.authorize.api.contract.v1.GetHostedProfilePageResponse;
import net.authorize.api.contract.v1.GetUnsettledTransactionListRequest;
import net.authorize.api.contract.v1.GetUnsettledTransactionListResponse;
import net.authorize.api.contract.v1.MerchantAuthenticationType;
import net.authorize.api.contract.v1.MessageTypeEnum;
import net.authorize.api.contract.v1.NameAndAddressType;
import net.authorize.api.contract.v1.OpaqueDataType;
import net.authorize.api.contract.v1.OrderType;
import net.authorize.api.contract.v1.Paging;
import net.authorize.api.contract.v1.PaymentProfile;
import net.authorize.api.contract.v1.PaymentType;
import net.authorize.api.contract.v1.SettingType;
import net.authorize.api.contract.v1.SolutionType;
import net.authorize.api.contract.v1.TransactionGroupStatusEnum;
import net.authorize.api.contract.v1.TransactionListOrderFieldEnum;
import net.authorize.api.contract.v1.TransactionListSorting;
import net.authorize.api.contract.v1.TransactionRequestType;
import net.authorize.api.contract.v1.TransactionResponse;
import net.authorize.api.contract.v1.TransactionTypeEnum;
import net.authorize.api.contract.v1.UpdateCustomerPaymentProfileRequest;
import net.authorize.api.contract.v1.UpdateCustomerPaymentProfileResponse;
import net.authorize.api.contract.v1.ValidationModeEnum;
import net.authorize.api.controller.CreateCustomerPaymentProfileController;
import net.authorize.api.controller.CreateCustomerProfileController;
import net.authorize.api.controller.CreateTransactionController;
import net.authorize.api.controller.DeleteCustomerPaymentProfileController;
import net.authorize.api.controller.GetCustomerProfileController;
import net.authorize.api.controller.GetHostedPaymentPageController;
import net.authorize.api.controller.GetHostedProfilePageController;
import net.authorize.api.controller.GetUnsettledTransactionListController;
import net.authorize.api.controller.UpdateCustomerPaymentProfileController;
import net.authorize.api.controller.base.ApiOperationBase;

/**
 * Processes a payment in the store for a given user's cart. This implementation processes a request through
 * Authorize.net's AIM integration method.
 * <p>
 * If the transaction is declined a result code of 10000 is returned, if it is successful a result code of 0 is
 * returned, and otherwise the result code sent back from Authorize.net is returned.
 * 
 * @author David Tobey
 * @see com.softslate.commerce.daos.payment.PaymentDAO
 */
public class AuthorizeNetAPIProcessor extends BasePaymentProcessor implements PaymentProcessor {

	static Log log = LogFactory.getLog(AuthorizeNetAIMProcessor.class);

	public Map process(Map parameters) throws Exception {
		log.info("Processing payment.");

		// AcceptHosted - the transaction is already processed directly by Authorize.net and we are relaying the
		// response from the iframe here. Just record the response.
		if (parameters.get("action") != null) {
			return processAcceptHostedResponse(parameters);
		}

		// Transaction type from settings
		String authorizeNetTxType = (String) getSettings().getValue("authorizeNetTxType");
		if (authorizeNetTxType.equals("AUTH_CAPTURE")) {
			parameters.put("TransactionType", TransactionTypeEnum.AUTH_CAPTURE_TRANSACTION.value());
		} else {
			parameters.put("TransactionType", TransactionTypeEnum.AUTH_ONLY_TRANSACTION.value());
		}
		return processTransaction(parameters);
	}

	public Map processAcceptHostedResponse(Map parameters) throws Exception {

		// Validate the response parameters.
		if (!parameters.get("action").equals("transactResponse") || parameters.get("response") == null) {
			return processErrorResponse(parameters, "badTransactResponse",
					"Invalid transaction response. Please contact an administrator to see if your payment was processed.");
		}

		// Check that the store is configured to use AcceptHosted.
		if (!getSettings().getValue("authorizeNetIntegrationMethod").startsWith("AcceptHosted")) {
			return processErrorResponse(parameters, "AcceptHostedNotActive",
					"AcceptHosted not active. Please contact an administrator to see if your payment was processed.");

		}

		// Create a CreateTransactionResponse object from the incoming JSON of the "response" parameter sent from
		// the iframe.
		Map responseAsMap = null;
		BigDecimal amount = null;
		try {
			ObjectMapper om = new ObjectMapper();
			TypeReference<HashMap<String, Object>> typeRef = new TypeReference<HashMap<String, Object>>() {
			};
			responseAsMap = om.readValue((String) parameters.get("response"), typeRef);
			amount = new BigDecimal((String) responseAsMap.get("totalAmount"));
		} catch (Exception e) {
			return processErrorResponse(parameters, "badTransactResponse",
					"Invalid transaction response. Please contact an administrator to see if your payment was processed.");
		}

		// Check that something in the response corresponds with the user's session...we have to use the customerId and
		// set it to something random each time - it's the only field that's part of both the request and response that
		// we can adjust.
		if (getUser().getCustomMap() == null || getUser().getCustomMap().get("AcceptHostedCustomerId") == null
				|| !getUser().getCustomMap().get("AcceptHostedCustomerId").equals(responseAsMap.get("customerId"))) {
			return processErrorResponse(parameters, "InvalidRefId",
					"Invalid reference id in the response. Please contact an administrator to see if your payment was processed.");
		} else {
			getUser().getCustomMap().remove("AcceptHostedCustomerId");
		}

		// Check the order total vs the response's transaction total.
		if (getUser().getOrder() == null || getUser().getOrder().getTotal() == null
				|| !getUser().getOrder().getTotal().equals(amount.doubleValue())) {
			return processErrorResponse(parameters, "InvalidOrderTotal",
					"Invalid order total. Please contact an administrator to see if your payment was processed.");
		}

		// Transaction type from settings
		if (parameters.get("TransactionType") == null) {
			String authorizeNetTxType = (String) getSettings().getValue("authorizeNetTxType");
			if (authorizeNetTxType.equals("AUTH_CAPTURE")) {
				parameters.put("TransactionType", TransactionTypeEnum.AUTH_CAPTURE_TRANSACTION.value());
			} else {
				parameters.put("TransactionType", TransactionTypeEnum.AUTH_ONLY_TRANSACTION.value());
			}
		}
		return processSuccessAcceptHostedResponse(parameters, responseAsMap);
	}

	public Map processAuthOnly(Map parameters) throws Exception {
		log.info("Processing auth capture.");

		// transaction fields
		parameters.put("TransactionType", TransactionTypeEnum.AUTH_ONLY_TRANSACTION.value());
		return processTransaction(parameters);
	}

	public Map processAuthCapture(Map parameters) throws Exception {
		log.info("Processing auth capture.");

		// transaction fields
		parameters.put("TransactionType", TransactionTypeEnum.AUTH_CAPTURE_TRANSACTION.value());
		return processTransaction(parameters);
	}

	public Map processPriorAuthCapture(Map parameters) throws Exception {
		log.info("Processing prior auth capture.");

		// transaction fields
		parameters.put("TransactionType", TransactionTypeEnum.PRIOR_AUTH_CAPTURE_TRANSACTION.value());
		return processTransaction(parameters);
	}

	public Map processVoid(Map parameters) throws Exception {
		log.info("Processing void.");

		// transaction fields
		parameters.put("TransactionType", TransactionTypeEnum.VOID_TRANSACTION.value());
		return processTransaction(parameters);
	}

	protected Map processTransaction(Map parameters) throws Exception {
		log.info("Processing transaction.");

		// Test or live mode
		Environment env = configureEnvironment(parameters);
		ApiOperationBase.setEnvironment(env);

		// Authentication
		ApiOperationBase.setMerchantAuthentication(configureMerchantAuthentication(parameters));

		// Create a new payment profile: If the customer already has an existing customer profile and wants to save the
		// payment information, we have to create a payment profile first. (There is no API call for creating just a
		// *payment* profile from a prior transaction, and if opaque data is being used, it can only be used once so we
		// have to use it to create the profile and then charge the profile).
		if ("1".equals(getSettings().getValue("authorizeNetEnableCustomerProfiles"))
				&& StringUtils.isBlank((String) parameters.get("paymentProfileID"))
				&& "true".equals(parameters.get("savePaymentInformation")) && getUser().getCustomer() != null
				&& StringUtils.isNotBlank(getUser().getCustomer().getCimProfileID())) {
			parameters.put("customerProfileID", getUser().getCustomer().getCimProfileID());
			Map createPaymentProfileResult = createPaymentProfile(parameters);
			String newPaymentProfileID = (String) createPaymentProfileResult.get("newPaymentProfileID");
			parameters.put("paymentProfileID", newPaymentProfileID); // Below we will charge the new profile
		}

		Map result = createTransation(parameters);
		return result;
	}

	protected Map createTransation(Map parameters) throws Exception {

		// Instantiate the transaction request
		TransactionRequestType txnRequest = initTransactionRequest(parameters);

		// Populate the solution ID
		txnRequest.setSolution(configureSolution(ApiOperationBase.getEnvironment()));

		if ("1".equals(getSettings().getValue("authorizeNetEnableCustomerProfiles"))
				&& StringUtils.isNotBlank((String) parameters.get("paymentProfileID"))) {

			// Paying with an existing paymentProfileID
			txnRequest.setProfile(configureProfileBeingUsedForPayment(parameters));
		} else {

			// Paying with a CC card, or opaque data from Accept.js
			txnRequest.setPayment(configurePayment(parameters));

			// If a new profile is being requested, indicate that
			txnRequest.setProfile(configureThatAProfileIsBeingRequested(parameters));

			// customer name and billing address fields - can't be sent with a paymentProfileID
			if (getUser() != null && getUser().getOrder() != null) {
				txnRequest.setBillTo(configureBillingAddress(parameters));
			}
		}

		// Total amount, tax and shipping
		configureTotals(txnRequest, parameters);

		// Invoice and customer reference fields
		configureReferenceFields(txnRequest, parameters);

		// Transaction settings such as email control fields and the duplicate transaction window
		configureTransactionSettings(txnRequest, parameters);

		// customer shipping fields
		if (getUser() != null && getUser().getOrder() != null && getUser().getOrder().firstDelivery() != null) {
			txnRequest.setShipTo(configureShippingAddress(parameters));
		}

		// Make the API Request
		CreateTransactionRequest apiRequest = new CreateTransactionRequest();
		apiRequest.setTransactionRequest(txnRequest);
		CreateTransactionController controller = new CreateTransactionController(apiRequest);
		controller.execute();
		CreateTransactionResponse response = controller.getApiResponse();

		// process the response
		return processResponse(response, parameters);
	}

	public Map createPaymentProfile(Map parameters) {

		// Test or live mode
		Environment env = configureEnvironment(parameters);
		ApiOperationBase.setEnvironment(env);

		// Authentication
		ApiOperationBase.setMerchantAuthentication(configureMerchantAuthentication(parameters));

		// Create the request
		CreateCustomerPaymentProfileRequest apiRequest = new CreateCustomerPaymentProfileRequest();
		String customerProfileID = null;
		if (StringUtils.isNotBlank((String) parameters.get("customerProfileID"))) {
			customerProfileID = (String) parameters.get("customerProfileID");
		} else {
			customerProfileID = getUser().getCustomer().getCimProfileID();
		}
		apiRequest.setCustomerProfileId(customerProfileID);
		apiRequest.setMerchantAuthentication(configureMerchantAuthentication(parameters));
		if (getUser().getOrder() != null) {
			apiRequest.setRefId(Integer.toString(getUser().getOrder().getOrderID()));
		}
		CustomerPaymentProfileType paymentProfileType = new CustomerPaymentProfileType();
		paymentProfileType.setBillTo(configureBillingAddress(parameters));
		paymentProfileType.setCustomerType(CustomerTypeEnum.INDIVIDUAL);
		paymentProfileType.setPayment(configurePayment(parameters));
		apiRequest.setPaymentProfile(paymentProfileType);

		// Send the request
		CreateCustomerPaymentProfileController controller = new CreateCustomerPaymentProfileController(apiRequest);
		controller.execute();
		CreateCustomerPaymentProfileResponse response = controller.getApiResponse();

		// Log the response and return the new paymentProfileID. Note: if a payment profile matching the card data
		// already exists, this will log an error (at INFO level), but it will also return the paymentProfileID we need.
		String resultCode = null;
		String messageCode = null;
		String resultMessage = null;
		if (response != null) {
			if (response.getMessages() != null) {
				MessageTypeEnum responseCode = response.getMessages().getResultCode();
				resultCode = response.getMessages().getResultCode().value();
				log.info("Create Payment Profile Result Code: " + responseCode);
				if (response.getMessages() != null && response.getMessages().getMessage() != null
						&& response.getMessages().getMessage().get(0) != null) {
					messageCode = response.getMessages().getMessage().get(0).getCode();
					resultMessage = response.getMessages().getMessage().get(0).getText();
					log.info("Create Payment Profile Message Code: " + resultCode);
					log.info("Create Payment Profile Message Text: " + resultMessage);
				}
			}
		}
		String newPaymentProfileID = response.getCustomerPaymentProfileId();
		log.info("newPaymentProfileID: " + newPaymentProfileID);
		Map result = new HashMap();
		result.put("resultCode", resultCode);
		result.put("messageCode", messageCode);
		result.put("resultMessage", resultMessage);
		result.put("newPaymentProfileID", newPaymentProfileID);
		return result;
	}

	protected Environment configureEnvironment(Map parameters) {
		String testMode = getSettings().getValue("authorizeNetTestRequest");
		if (testMode != null && testMode.equals("1")) {
			return Environment.SANDBOX;
		} else {
			return Environment.PRODUCTION;
		}
	}

	protected MerchantAuthenticationType configureMerchantAuthentication(Map parameters) {
		MerchantAuthenticationType merchantAuthenticationType = new MerchantAuthenticationType();
		merchantAuthenticationType.setName(getSettings().getValue("authorizeNetLoginID"));
		merchantAuthenticationType.setTransactionKey(getSettings().getValue("authorizeNetTxKey"));
		return merchantAuthenticationType;
	}

	protected TransactionRequestType initTransactionRequest(Map parameters) {
		TransactionRequestType txnRequest = new TransactionRequestType();
		txnRequest.setCurrencyCode(getSettings().getValue("currencyCode"));
		txnRequest.setTransactionType((String) parameters.get("TransactionType"));

		// For voids or other transactions requiring a reference to a prior transaction
		txnRequest.setRefTransId((String) parameters.get("x_trans_id"));
		return txnRequest;
	}

	protected SolutionType configureSolution(Environment env) {
		SolutionType solutionType = new SolutionType();
		if (env == Environment.SANDBOX) {
			solutionType.setId("AAA100302");
		} else {
			solutionType.setId("AAA172581");
		}
		return solutionType;
	}

	protected CustomerProfilePaymentType configureProfileBeingUsedForPayment(Map parameters) {
		CustomerProfilePaymentType profileType = new CustomerProfilePaymentType();
		String customerProfileID = null;
		if (StringUtils.isNotBlank((String) parameters.get("customerProfileID"))) {
			customerProfileID = (String) parameters.get("customerProfileID");
		} else {
			customerProfileID = getUser().getCustomer().getCimProfileID();
		}
		String paymentProfileID = (String) parameters.get("paymentProfileID");
		String creditCardSecurityCode = (String) parameters.get("creditCardSecurityCode");
		profileType.setCustomerProfileId(customerProfileID);
		PaymentProfile paymentProfile = new PaymentProfile();
		paymentProfile.setPaymentProfileId(paymentProfileID);
		if (creditCardSecurityCode != null && creditCardSecurityCode.length() > 0) {
			paymentProfile.setCardCode(creditCardSecurityCode);
		}
		profileType.setPaymentProfile(paymentProfile);
		return profileType;
	}

	protected CustomerProfilePaymentType configureThatAProfileIsBeingRequested(Map parameters) {
		CustomerProfilePaymentType profileType = new CustomerProfilePaymentType();
		if (getUser().getCustomer() != null && StringUtils.isBlank(getUser().getCustomer().getCimProfileID())
				&& "1".equals(getSettings().getValue("authorizeNetEnableCustomerProfiles"))
				&& "true".equals(parameters.get("savePaymentInformation"))) {
			profileType.setCreateProfile(Boolean.TRUE);
		} else {
			profileType.setCreateProfile(Boolean.FALSE);
		}
		return profileType;
	}

	protected PaymentType configurePayment(Map parameters) {
		PaymentType paymentType = new PaymentType();
		if (StringUtils.isNotBlank((String) parameters.get("acceptJsDataDesc"))
				&& StringUtils.isNotBlank((String) parameters.get("acceptJsDataValue"))) {

			// Use opaque data from Accept.js
			OpaqueDataType opaqueData = new OpaqueDataType();
			opaqueData.setDataDescriptor((String) parameters.get("acceptJsDataDesc"));
			opaqueData.setDataValue((String) parameters.get("acceptJsDataValue"));
			paymentType.setOpaqueData(opaqueData);
		} else {

			// Use credit card data if opaque data is not present
			CreditCardType creditCard = new CreditCardType();
			creditCard.setCardNumber((String) parameters.get("decryptedNumber"));
			if (parameters.get("creditCardExpirationMonth") != null
					&& parameters.get("creditCardExpirationYear") != null) {
				creditCard.setExpirationDate(parameters.get("creditCardExpirationMonth").toString()
						+ parameters.get("creditCardExpirationYear").toString());
			}
			if (parameters.get("creditCardSecurityCode") != null) {
				creditCard.setCardCode((String) parameters.get("creditCardSecurityCode"));
			}
			paymentType.setCreditCard(creditCard);
		}
		return paymentType;
	}

	protected void configureTotals(TransactionRequestType txnRequest, Map parameters) {
		if (getUser() != null && getUser().getOrder() != null && parameters.get("x_amount") == null
				&& getUser().getOrder().getTotal() != null) {
			Double total = getUser().getOrder().getTotal();
			total = utils().roundForMoney(total);
			parameters.put("amount", new BigDecimal(total).setScale(2, RoundingMode.HALF_UP));
			txnRequest.setAmount(new BigDecimal(total).setScale(2, RoundingMode.HALF_UP));
		} else if (parameters.get("x_amount") != null) {
			parameters.put("amount",
					new BigDecimal((String) parameters.get("x_amount")).setScale(2, RoundingMode.HALF_UP));
			txnRequest.setAmount(new BigDecimal((String) parameters.get("x_amount")).setScale(2, RoundingMode.HALF_UP));
		}
		ExtendedAmountType freight = new ExtendedAmountType();
		freight.setName("Freight");
		if (getUser() != null && getUser().getOrder() != null && parameters.get("x_freight") == null
				&& getUser().getOrder().getShipping() != null) {
			Double shipping = getUser().getOrder().getShipping();
			shipping = utils().roundForMoney(shipping);
			freight.setAmount(new BigDecimal(shipping).setScale(2, RoundingMode.HALF_UP));
			txnRequest.setShipping(freight);
		} else if (parameters.get("x_freight") != null) {
			freight.setAmount(new BigDecimal((String) parameters.get("x_freight")).setScale(2, RoundingMode.HALF_UP));
			txnRequest.setShipping(freight);
		}
		ExtendedAmountType tax = new ExtendedAmountType();
		tax.setName("tax");
		if (getUser() != null && getUser().getOrder() != null && parameters.get("x_tax") == null
				&& getUser().getOrder().getTax() != null) {
			Double taxAmount = getUser().getOrder().getTax();
			taxAmount = utils().roundForMoney(taxAmount);
			tax.setAmount(new BigDecimal(taxAmount).setScale(2, RoundingMode.HALF_UP));
			txnRequest.setTax(tax);
		} else if (parameters.get("x_tax") != null) {
			tax.setAmount(new BigDecimal((String) parameters.get("x_tax")).setScale(2, RoundingMode.HALF_UP));
			txnRequest.setTax(tax);
		}
	}

	protected void configureReferenceFields(TransactionRequestType txnRequest, Map parameters) {
		txnRequest.setPoNumber(escapeForIFrameCommunicator((String) parameters.get("poNumber")));
		OrderType order = new OrderType();
		order.setInvoiceNumber(escapeForIFrameCommunicator((String) parameters.get("invoiceNumber")));
		order.setDescription(escapeForIFrameCommunicator((String) parameters.get("description")));
		txnRequest.setOrder(order);
		CustomerDataType customer = new CustomerDataType();
		if (getUser().getCustomer() != null) {
			customer.setId(Integer.toString(getUser().getCustomer().getCustomerID()));
		} else {
			customer.setId((String) parameters.get("customerID"));
		}
		txnRequest.setCustomer(customer);
		txnRequest.setCustomerIP((String) parameters.get("ipAddress"));
	}

	protected void configureTransactionSettings(TransactionRequestType txnRequest, Map parameters) {
		ArrayOfSetting settings = new ArrayOfSetting();
		SettingType emailCustomer = new SettingType();
		emailCustomer.setSettingName("emailCustomer");
		String emailCust = (String) getSettings().getValue("authorizeNetEmailCustomer");
		if (emailCust != null && emailCust.equals("1")) {
			emailCustomer.setSettingValue("TRUE");
		} else {
			emailCustomer.setSettingValue("FALSE");
		}
		settings.getSetting().add(emailCustomer);
		txnRequest.setTransactionSettings(settings);

		// duplicateWindow
		SettingType duplicateWindow = new SettingType();
		duplicateWindow.setSettingName("duplicateWindow");
		duplicateWindow.setSettingValue(getSettings().getValue("authorizeNetDupeTxTimeout"));
		settings.getSetting().add(duplicateWindow);
	}

	protected CustomerAddressType configureBillingAddress(Map parameters) {
		CustomerAddressType billTo = new CustomerAddressType();
		if (parameters.get("billingFirstName") != null) {
			billTo.setFirstName(escapeForIFrameCommunicator((String) parameters.get("billingFirstName")));
		} else if (getUser().getOrder() != null && getUser().getOrder().getFirstName() != null) {
			billTo.setFirstName(escapeForIFrameCommunicator(getUser().getOrder().getFirstName()));
		}
		if (parameters.get("billingLastName") != null) {
			billTo.setLastName(escapeForIFrameCommunicator((String) parameters.get("billingLastName")));
		} else if (getUser().getOrder() != null && getUser().getOrder().getLastName() != null) {
			billTo.setLastName(escapeForIFrameCommunicator(getUser().getOrder().getLastName()));
		}
		if (parameters.get("billingOrganization") != null) {
			billTo.setCompany(escapeForIFrameCommunicator((String) parameters.get("billingOrganization")));
		} else if (getUser().getOrder() != null && getUser().getOrder().getOrganization() != null) {
			billTo.setCompany(escapeForIFrameCommunicator(getUser().getOrder().getOrganization()));
		}
		if (parameters.get("billingAddress1") != null) {
			billTo.setAddress(escapeForIFrameCommunicator((String) parameters.get("billingAddress1")));
		} else if (getUser().getOrder() != null && getUser().getOrder().getAddress1() != null) {
			billTo.setAddress(escapeForIFrameCommunicator(getUser().getOrder().getAddress1()));
		}
		if (parameters.get("billingAddress2") != null) {
			billTo.setAddress(escapeForIFrameCommunicator(parameters.get("billingAddress1") + " " + parameters.get("billingAddress2")));
		} else if (getUser().getOrder() != null && getUser().getOrder().getAddress2() != null) {
			billTo.setAddress(escapeForIFrameCommunicator(getUser().getOrder().getAddress1() + " " + getUser().getOrder().getAddress2()));
		}
		if (parameters.get("billingCity") != null) {
			billTo.setCity(escapeForIFrameCommunicator((String) parameters.get("billingCity")));
		} else if (getUser().getOrder() != null && getUser().getOrder().getCity() != null) {
			billTo.setCity(escapeForIFrameCommunicator(getUser().getOrder().getCity()));
		}
		if (parameters.get("billingState") != null) {
			billTo.setState(escapeForIFrameCommunicator((String) parameters.get("billingState")));
		} else if (getUser().getOrder() != null && getUser().getOrder().getState() != null) {
			billTo.setState(escapeForIFrameCommunicator(getUser().getOrder().getState()));
		}
		if (parameters.get("billingPostalCode") != null) {
			billTo.setZip(escapeForIFrameCommunicator((String) parameters.get("billingPostalCode")));
		} else if (getUser().getOrder() != null && getUser().getOrder().getPostalCode() != null) {
			billTo.setZip(escapeForIFrameCommunicator(getUser().getOrder().getPostalCode()));
		}
		if (parameters.get("billingCountry") != null) {
			billTo.setCountry(escapeForIFrameCommunicator((String) parameters.get("billingCountry")));
		} else if (getUser().getOrder() != null && getUser().getOrder().getCountry() != null) {
			billTo.setCountry(escapeForIFrameCommunicator(getUser().getOrder().getCountry()));
		}
		if (parameters.get("billingPhone1") != null) {
			billTo.setPhoneNumber((String) parameters.get("billingPhone1"));
		} else if (getUser().getOrder() != null && getUser().getOrder().getPhone1() != null) {
			billTo.setPhoneNumber(getUser().getOrder().getPhone1());
		}
		if (parameters.get("billingPhone2") != null) {
			billTo.setFaxNumber(escapeForIFrameCommunicator((String) parameters.get("billingPhone2")));
		} else if (getUser().getOrder() != null && getUser().getOrder().getPhone2() != null) {
			billTo.setFaxNumber(escapeForIFrameCommunicator(getUser().getOrder().getPhone2()));
		}
		if (parameters.get("billingEmail1") != null) {
			billTo.setEmail(escapeForIFrameCommunicator((String) parameters.get("billingEmail1")));
		} else if (getUser().getOrder() != null && getUser().getOrder().getEmail1() != null) {
			billTo.setEmail(escapeForIFrameCommunicator(getUser().getOrder().getEmail1()));
		}
		return billTo;
	}

	protected NameAndAddressType configureShippingAddress(Map parameters) {
		CustomerAddressType shipTo = new CustomerAddressType();
		if (getUser().getOrder().firstDelivery().getFirstName() != null) {
			shipTo.setFirstName(escapeForIFrameCommunicator(getUser().getOrder().firstDelivery().getFirstName()));
		}
		if (getUser().getOrder().firstDelivery().getLastName() != null) {
			shipTo.setLastName(escapeForIFrameCommunicator(getUser().getOrder().firstDelivery().getLastName()));
		}
		if (getUser().getOrder().firstDelivery().getOrganization() != null) {
			shipTo.setCompany(escapeForIFrameCommunicator(getUser().getOrder().firstDelivery().getOrganization()));
		}
		if (getUser().getOrder().firstDelivery().getAddress1() != null) {
			shipTo.setAddress(escapeForIFrameCommunicator(getUser().getOrder().firstDelivery().getAddress1()));
		}
		if (getUser().getOrder().firstDelivery().getAddress2() != null) {
			shipTo.setAddress(escapeForIFrameCommunicator(getUser().getOrder().firstDelivery().getAddress1() + " "
					+ getUser().getOrder().firstDelivery().getAddress2()));
		}
		if (getUser().getOrder().firstDelivery().getCity() != null) {
			shipTo.setCity(escapeForIFrameCommunicator(getUser().getOrder().firstDelivery().getCity()));
		}
		if (getUser().getOrder().firstDelivery().getState() != null) {
			shipTo.setState(escapeForIFrameCommunicator(getUser().getOrder().firstDelivery().getState()));
		}
		if (getUser().getOrder().firstDelivery().getPostalCode() != null) {
			shipTo.setZip(escapeForIFrameCommunicator(getUser().getOrder().firstDelivery().getPostalCode()));
		}
		if (getUser().getOrder().firstDelivery().getCountry() != null) {
			shipTo.setCountry(escapeForIFrameCommunicator(getUser().getOrder().firstDelivery().getCountry()));
		}
		if (getUser().getOrder().firstDelivery().getPhone1() != null) {
			shipTo.setPhoneNumber(escapeForIFrameCommunicator(getUser().getOrder().firstDelivery().getPhone1()));
		}
		if (getUser().getOrder().firstDelivery().getPhone2() != null) {
			shipTo.setFaxNumber(escapeForIFrameCommunicator(getUser().getOrder().firstDelivery().getPhone2()));
		}
		if (getUser().getOrder().firstDelivery().getEmail1() != null) {
			shipTo.setEmail(escapeForIFrameCommunicator(getUser().getOrder().firstDelivery().getEmail1()));
		}
		return shipTo;
	}

	/*
	 * The Authorize.net AcceptHosted response will send the address and other values back to the
	 * IFrameCommunicator.html file, which in turn will relay them to /CheckoutPayment.do. The JS in IFrameCommunicator
	 * needs to parse the Authorize.net response, which includes splitting on amper's ands (&) and equals signs (=). So
	 * we have to make sure those characters are not in the data we send them in the first place. Authorize.net escapes
	 * other characters on their side but not & or =.
	 */
	protected String escapeForIFrameCommunicator(String value) {
		if (value == null) {
			return null;
		}
		value = value.replaceAll("&", "and");
		value = value.replaceAll("=", "equals");
		return value;
	}

	protected Map processResponse(CreateTransactionResponse response, Map parameters) throws Exception {
		String resultCode = null;
		String messageCode = null;
		String resultMessage = null;
		if (response != null) {

			// Log the overall response and check it for an error (such as an authentication issue)
			if (response.getMessages() != null) {

				// Log it
				resultCode = response.getMessages().getResultCode().value();
				MessageTypeEnum responseCode = response.getMessages().getResultCode();
				log.info("Result Code: " + responseCode);
				if (response.getMessages() != null && response.getMessages().getMessage() != null
						&& response.getMessages().getMessage().get(0) != null) {
					messageCode = response.getMessages().getMessage().get(0).getCode();
					resultMessage = response.getMessages().getMessage().get(0).getText();
					log.info("Result Code: " + resultCode);
					log.info("Message Code: " + messageCode);
					log.info("Message Text: " + resultMessage);
				}

				// Check it
				if (responseCode != null && responseCode == MessageTypeEnum.ERROR) {
					return processErrorResponse(parameters, messageCode, resultMessage);
				}
			}

			// Log the transaction response and check it for an error (such as payment declined)
			TransactionResponse transResponse = response.getTransactionResponse();
			if (transResponse != null) {

				// Log it
				log.info("Transaction ID: " + transResponse.getTransId());
				log.info("Response Code: " + transResponse.getResponseCode());
				if (transResponse.getMessages() != null && transResponse.getMessages().getMessage() != null
						&& transResponse.getMessages().getMessage().get(0) != null) {
					resultCode = transResponse.getMessages().getMessage().get(0).getCode();
					resultMessage = transResponse.getMessages().getMessage().get(0).getDescription();
					log.info("Transaction Response Code: " + resultCode);
					log.info("Transaction Response Description: " + resultMessage);
				}
				if (transResponse.getErrors() != null && transResponse.getErrors().getError() != null
						&& transResponse.getErrors().getError().get(0) != null) {
					resultCode = transResponse.getErrors().getError().get(0).getErrorCode();
					resultMessage = transResponse.getErrors().getError().get(0).getErrorText();
					log.info("Transaction Error Code: " + resultCode);
					log.info("Transaction Error Description: " + resultMessage);
				}

				// Check it
				if (transResponse.getResponseCode() != null && transResponse.getResponseCode().equals("1")) {
					log.info("Payment Approved");
					return processSuccessResponse(parameters, response);
				} else if (transResponse.getResponseCode() != null && transResponse.getResponseCode().equals("2")) {
					log.info("Payment Declined");

					// The Transaction Error Description is very misleading in this case - tells the user to call
					// Merchant Services. Just tell them the card was declined.
					return processErrorResponse(parameters, resultCode, "Credit Card Declined");
				} else if (transResponse.getResponseCode() != null && transResponse.getResponseCode().equals("3")) {
					log.info("Payment Error");
					return processErrorResponse(parameters, resultCode, resultMessage);
				} else if (transResponse.getResponseCode() != null && transResponse.getResponseCode().equals("4")) {
					log.info("Held for Review - treating as successful");
					return processSuccessResponse(parameters, response);
				} else {
					log.info("Not a recognized transaction response code - treating as a failure: "
							+ transResponse.getResponseCode());
					return processErrorResponse(parameters, resultCode, resultMessage);
				}
			} else {
				log.info("Null Transaction Response - treating as a failure");
				return processErrorResponse(parameters, resultCode, resultMessage);
			}
		} else {
			log.info("Null Response - treating as a failure");
			return processErrorResponse(parameters, resultCode, resultMessage);
		}
	}

	protected Map processSuccessResponse(Map parameters, CreateTransactionResponse response) throws Exception {
		Map result = new HashMap();
		result.put("resultCode", "0");
		result.put("resultMessage",
				response.getTransactionResponse().getMessages().getMessage().get(0).getDescription());
		Payment payment = (Payment) getBusinessObjectFactory().createObject("paymentImplementer");
		BeanUtils.copyProperties(payment, parameters);
		payment.setStatus("Received");
		payment.setCreated(formatDateTime(new java.util.Date()));
		payment.setProcessorClassName(this.getClass().getName());
		String creditCardNumberDisplay = null;
		String decryptedNumber = payment.getDecryptedNumber();
		if (decryptedNumber != null && decryptedNumber.length() > 4) {
			creditCardNumberDisplay = "XXXXXXXXXXXX"
					+ decryptedNumber.substring(decryptedNumber.length() - 4, decryptedNumber.length());
			payment.setCreditCardNumberDisplay(creditCardNumberDisplay);
		}
		payment.setValueFieldLabels(
				"TRANSACTION_TYPE,AMOUNT,CARD_CODE,RESPONSE_CODE,AVS_CODE,AUTHCODE,TRANSACTION_ID,NONCE");
		payment.setValue1((String) parameters.get("TransactionType"));
		if (parameters.get("amount") != null) {
			payment.setAmount(((BigDecimal) parameters.get("amount")).doubleValue());
			payment.setValue2(((BigDecimal) parameters.get("amount")).toString());
		}
		payment.setValue3(response.getTransactionResponse().getCavvResultCode());
		payment.setValue4(response.getTransactionResponse().getResponseCode());
		payment.setValue5(response.getTransactionResponse().getAvsResultCode());
		payment.setValue6(response.getTransactionResponse().getAuthCode());
		payment.setValue7(response.getTransactionResponse().getTransId());

		// Save the payment profile ID used for the payment, including if a new customer profile was just created
		if (StringUtils.isNotBlank((String) parameters.get("paymentProfileID"))) {
			payment.setPaymentProfileID((String) parameters.get("paymentProfileID"));
		} else if (response.getProfileResponse() != null
				&& response.getProfileResponse().getCustomerPaymentProfileIdList() != null
				&& response.getProfileResponse().getCustomerPaymentProfileIdList().getNumericString() != null
				&& response.getProfileResponse().getCustomerPaymentProfileIdList().getNumericString().size() > 0) {
			payment.setPaymentProfileID(
					response.getProfileResponse().getCustomerPaymentProfileIdList().getNumericString().get(0));
		}

		// Save the Accept.js nonce
		if (StringUtils.isNotBlank((String) parameters.get("acceptJsDataValue"))) {
			payment.setValue8((String) parameters.get("acceptJsDataValue"));
		}

		// If a customer profile was used, save the customer profile id in Order.paymentProfileID
		if (response.getProfileResponse() != null && response.getProfileResponse().getCustomerProfileId() != null) {
			getUser().getOrder().setPaymentProfileID(response.getProfileResponse().getCustomerProfileId());

			// If the Customer chose to save the info, save the customer profile id in the customer record
			if ("true".equals(parameters.get("savePaymentInformation")) && getUser().getCustomer() != null) {
				CustomerProcessor customerProcessor = getBusinessObjectFactory().createObject(CustomerProcessor.class);
				customerProcessor.updateCustomer(
						Collections.singletonMap("cimProfileID", response.getProfileResponse().getCustomerProfileId()));
				getUser().getCustomer().setCimProfileID(response.getProfileResponse().getCustomerProfileId());
			}
		}
		payment.setOrder(getUser().getOrder());
		if (getUser().getOrder().getPayments() == null)
			getUser().getOrder().setPayments(new LinkedHashSet());
		getUser().getOrder().getPayments().add(payment);
		PaymentDAO paymentDAO = (PaymentDAO) getDaoFactory().createDAO("paymentDAOImplementer");
		paymentDAO.setPayment(payment);
		paymentDAO.insertPayment(false);
		result.put("newPayment", payment);
		return result;
	}

	protected Map processSuccessAcceptHostedResponse(Map parameters, Map responseAsMap) throws Exception {
		Map result = new HashMap();
		result.put("resultCode", "0");
		result.put("resultMessage", "AcceptHosted response.");
		Payment payment = (Payment) getBusinessObjectFactory().createObject("paymentImplementer");
		BeanUtils.copyProperties(payment, parameters);
		payment.setStatus("Received");
		payment.setCreated(formatDateTime(new java.util.Date()));
		payment.setProcessorClassName(this.getClass().getName());
		payment.setCreditCardNumberDisplay((String) responseAsMap.get("accountNumber"));
		payment.setCreditCardType((String) responseAsMap.get("accountType"));
		payment.setValueFieldLabels(
				"TRANSACTION_TYPE,AMOUNT,CARD_CODE,RESPONSE_CODE,AVS_CODE,AUTHCODE,TRANSACTION_ID,NONCE");
		payment.setValue1((String) parameters.get("TransactionType"));
		if (responseAsMap.get("totalAmount") != null) {
			BigDecimal amount = new BigDecimal((String) responseAsMap.get("totalAmount"));
			payment.setAmount(amount.doubleValue());
			payment.setValue2(amount.toString());
		}

		// Note: we can't record the paymentProfileID if any, or the AVS or CAVV responses, because we don't get them
		// back.
		payment.setValue4((String) responseAsMap.get("responseCode"));
		payment.setValue6((String) responseAsMap.get("authorization"));
		payment.setValue7((String) responseAsMap.get("transId"));

		// Put the entire response in the text field, value9, for reference.
		payment.setValue9(responseAsMap.toString());
		payment.setOrder(getUser().getOrder());
		if (getUser().getOrder().getPayments() == null)
			getUser().getOrder().setPayments(new LinkedHashSet());
		getUser().getOrder().getPayments().add(payment);
		PaymentDAO paymentDAO = (PaymentDAO) getDaoFactory().createDAO("paymentDAOImplementer");
		paymentDAO.setPayment(payment);
		paymentDAO.insertPayment(false);
		result.put("newPayment", payment);
		return result;
	}

	protected Map processErrorResponse(Map parameters, String code, String message) throws Exception {
		Map result = new HashMap();
		if (code == null) {
			code = "10000";
		}
		result.put("resultCode", code);
		result.put("resultMessage", message);
		return result;
	}

	public Map retrieveUnsettledTransactionList(int limit, int offset) {

		// Test or live mode
		Environment env = configureEnvironment(null);
		ApiOperationBase.setEnvironment(env);

		// Authentication
		ApiOperationBase.setMerchantAuthentication(configureMerchantAuthentication(null));

		// Create the request
		GetUnsettledTransactionListRequest request = new GetUnsettledTransactionListRequest();
		request.setStatus(TransactionGroupStatusEnum.ANY);
		Paging paging = new Paging();
		paging.setLimit(limit);
		paging.setOffset(offset);
		request.setPaging(paging);
		TransactionListSorting sorting = new TransactionListSorting();
		sorting.setOrderBy(TransactionListOrderFieldEnum.SUBMIT_TIME_UTC);
		sorting.setOrderDescending(true);
		request.setSorting(sorting);

		// Send the request
		GetUnsettledTransactionListController controller = new GetUnsettledTransactionListController(request);
		controller.execute();
		GetUnsettledTransactionListResponse response = controller.getApiResponse();
		String resultCode = null;
		String messageCode = null;
		String resultMessage = null;
		if (response != null) {
			log.info("GetUnsettledTransactionList Number of results: " + response.getTotalNumInResultSet());
			if (response.getMessages() != null) {
				resultCode = response.getMessages().getResultCode().value();
				MessageTypeEnum responseCode = response.getMessages().getResultCode();
				log.info("GetUnsettledTransactionList Result Code: " + responseCode);
				if (response.getMessages() != null && response.getMessages().getMessage() != null
						&& response.getMessages().getMessage().get(0) != null) {
					messageCode = response.getMessages().getMessage().get(0).getCode();
					resultMessage = response.getMessages().getMessage().get(0).getText();
					log.info("GetUnsettledTransactionList resultCode: " + resultCode);
					log.info("GetUnsettledTransactionList Message Code: " + messageCode);
					log.info("GetUnsettledTransactionList Message Text: " + resultMessage);
				}
			}
		}
		Map result = new HashMap();
		result.put("resultCode", resultCode);
		result.put("messageCode", messageCode);
		result.put("resultMessage", resultMessage);
		result.put("transactions", response.getTransactions());
		return result;
	}

	public Map createCustomerProfile(Customer customer, Map parameters) throws Exception {

		// Test or live mode
		Environment env = configureEnvironment(parameters);
		ApiOperationBase.setEnvironment(env);

		// Authentication
		ApiOperationBase.setMerchantAuthentication(configureMerchantAuthentication(parameters));

		// Create the request
		CreateCustomerProfileRequest apiRequest = new CreateCustomerProfileRequest();
		apiRequest.setMerchantAuthentication(configureMerchantAuthentication(null));
		apiRequest.setRefId("c-" + Integer.toString(customer.getCustomerID()));

		CustomerProfileType customerProfile = new CustomerProfileType();
		customerProfile.setMerchantCustomerId("c-" + Integer.toString(customer.getCustomerID()));
		customerProfile.setEmail(StringUtils.stripToEmpty(customer.getEmail1()));
		customerProfile.setDescription(StringUtils.stripToEmpty(customer.getUserName()));
		apiRequest.setProfile(customerProfile);

		// Send the request
		CreateCustomerProfileController controller = new CreateCustomerProfileController(apiRequest);
		controller.execute();
		CreateCustomerProfileResponse response = controller.getApiResponse();

		// Log the response and return the new paymentProfileID. Note: if a payment profile matching the card data
		// already exists, this will log an error (at INFO level), but it will also return the paymentProfileID we need.
		String resultCode = null;
		String messageCode = null;
		String resultMessage = null;
		if (response != null) {
			if (response.getMessages() != null) {
				MessageTypeEnum responseCode = response.getMessages().getResultCode();
				resultCode = response.getMessages().getResultCode().value();
				log.info("Create Payment Profile Result Code: " + responseCode);
				if (response.getMessages() != null && response.getMessages().getMessage() != null
						&& response.getMessages().getMessage().get(0) != null) {
					messageCode = response.getMessages().getMessage().get(0).getCode();
					resultMessage = response.getMessages().getMessage().get(0).getText();
					log.info("Create Payment Profile Message Code: " + resultCode);
					log.info("Create Payment Profile Message Text: " + resultMessage);
				}
			}
		}
		String newProfileID = response.getCustomerProfileId();
		log.info("newProfileID: " + newProfileID);

		// If the Customer chose to save the info, save the customer profile id in the customer record
		if (newProfileID != null && newProfileID.length() > 0) {
			CustomerProcessor customerProcessor = getBusinessObjectFactory().createObject(CustomerProcessor.class);
			customerProcessor.updateCustomer(Collections.singletonMap("cimProfileID", newProfileID));
		}
		Map result = new HashMap();
		result.put("resultCode", resultCode);
		result.put("messageCode", messageCode);
		result.put("resultMessage", resultMessage);
		result.put("newProfileID", newProfileID);
		return result;
	}

	public List<CustomerPaymentProfileMaskedType> retrievePaymentProfiles(String cimProfileID, boolean includeExpired) {
		List<CustomerPaymentProfileMaskedType> paymentProfiles = new ArrayList<CustomerPaymentProfileMaskedType>();

		// Test or live mode
		Environment env = configureEnvironment(null);
		ApiOperationBase.setEnvironment(env);

		// Authentication
		ApiOperationBase.setMerchantAuthentication(configureMerchantAuthentication(null));

		// Create the request
		GetCustomerProfileRequest apiRequest = new GetCustomerProfileRequest();
		apiRequest.setCustomerProfileId(cimProfileID);
		apiRequest.setUnmaskExpirationDate(Boolean.TRUE);

		GetCustomerProfileController controller = new GetCustomerProfileController(apiRequest);
		controller.execute();
		GetCustomerProfileResponse response = controller.getApiResponse();

		// Log the response and return the new payment profiles.
		String resultCode = null;
		String messageCode = null;
		String resultMessage = null;
		if (response != null) {
			if (response.getMessages() != null) {
				MessageTypeEnum responseCode = response.getMessages().getResultCode();
				resultCode = response.getMessages().getResultCode().value();
				log.info("Create Payment Profile Result Code: " + responseCode);
				if (response.getMessages() != null && response.getMessages().getMessage() != null
						&& response.getMessages().getMessage().get(0) != null) {
					messageCode = response.getMessages().getMessage().get(0).getCode();
					resultMessage = response.getMessages().getMessage().get(0).getText();
					log.info("Retrieve Payment Profiles resultCode: " + resultCode);
					log.info("Retrieve Payment Profiles Message Code: " + messageCode);
					log.info("Retrieve Payment Profiles Message Text: " + resultMessage);
				}
			}
		}

		// Collect the payment profiles and return them.
		if (response.getProfile() != null && response.getProfile().getPaymentProfiles() != null) {
			for (CustomerPaymentProfileMaskedType profile : response.getProfile().getPaymentProfiles()) {
				if (profile.getPayment() != null && profile.getPayment().getCreditCard() != null) {
					if (includeExpired) {
						paymentProfiles.add(profile);
					} else if (profile.getPayment().getCreditCard().getExpirationDate() != null) {

						// If called for, filter out expired credit cards
						String expirationDateStr = profile.getPayment().getCreditCard().getExpirationDate();
						try {
							SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM", Locale.getDefault());
							Date expirationDate = formatter.parse(expirationDateStr);
							Calendar now = Calendar.getInstance(getUser().getLocale());
							Calendar cal = Calendar.getInstance(getUser().getLocale());
							cal.setTime(expirationDate);
							cal.add(Calendar.MONTH, 1);
							if (cal.before(now)) {
								log.info("Skipping expired credit card profile: " + expirationDateStr);
							} else {
								paymentProfiles.add(profile);
							}
						} catch (ParseException e) {
							log.error("Error parsing dateTime: " + expirationDateStr, e);
						}
					}
				} else if (profile.getPayment() != null) {
					paymentProfiles.add(profile);
				}
			}
			log.info("Found existing payment profiles: " + paymentProfiles.size());
		}
		return paymentProfiles;
	}

	public CustomerPaymentProfileMaskedType retrievePaymentProfile(String cimProfileID, String cimPaymentProfileID)
			throws Exception {
		CustomerPaymentProfileMaskedType paymentProfile = null;
		List<CustomerPaymentProfileMaskedType> profiles = retrievePaymentProfiles(cimProfileID, true);
		for (CustomerPaymentProfileMaskedType thisProfile : profiles) {
			if (thisProfile.getCustomerPaymentProfileId().equals(cimPaymentProfileID)) {
				paymentProfile = thisProfile;
				break;
			}
		}
		return paymentProfile;
	}

	public void copyPaymentPaymentProfileToAddressForm(CustomerPaymentProfileMaskedType paymentProfile,
			BillingAddressForm form) {
		if (paymentProfile.getBillTo() != null) {
			form.setBillingAddress1(paymentProfile.getBillTo().getAddress());
			form.setBillingCity(paymentProfile.getBillTo().getCity());
			form.setBillingOrganization(paymentProfile.getBillTo().getCompany());
			form.setBillingCountry(paymentProfile.getBillTo().getCountry());
			form.setBillingFirstName(paymentProfile.getBillTo().getFirstName());
			form.setBillingLastName(paymentProfile.getBillTo().getLastName());
			form.setBillingPhone1(paymentProfile.getBillTo().getPhoneNumber());
			form.setBillingState(paymentProfile.getBillTo().getState());
			form.setBillingPostalCode(paymentProfile.getBillTo().getZip());
		}
	}

	public Map editPaymentProfile(Map parameters) {

		// Test or live mode
		Environment env = configureEnvironment(parameters);
		ApiOperationBase.setEnvironment(env);

		// Authentication
		ApiOperationBase.setMerchantAuthentication(configureMerchantAuthentication(parameters));

		// Create the request
		UpdateCustomerPaymentProfileRequest apiRequest = new UpdateCustomerPaymentProfileRequest();
		String customerProfileID = null;
		if (StringUtils.isNotBlank((String) parameters.get("customerProfileID"))) {
			customerProfileID = (String) parameters.get("customerProfileID");
		} else {
			customerProfileID = getUser().getCustomer().getCimProfileID();
		}
		apiRequest.setCustomerProfileId(customerProfileID);
		apiRequest.setMerchantAuthentication(configureMerchantAuthentication(parameters));
		if (getUser().getOrder() != null) {
			apiRequest.setRefId(Integer.toString(getUser().getOrder().getOrderID()));
		}
		CustomerPaymentProfileExType paymentProfileType = new CustomerPaymentProfileExType();
		paymentProfileType.setCustomerPaymentProfileId((String) parameters.get("paymentProfileID"));
		paymentProfileType.setBillTo(configureBillingAddress(parameters));
		paymentProfileType.setCustomerType(CustomerTypeEnum.INDIVIDUAL);
		paymentProfileType.setPayment(configurePayment(parameters));
		apiRequest.setPaymentProfile(paymentProfileType);
		apiRequest.setValidationMode(ValidationModeEnum.NONE);

		// Send the request
		UpdateCustomerPaymentProfileController controller = new UpdateCustomerPaymentProfileController(apiRequest);
		controller.execute();
		UpdateCustomerPaymentProfileResponse response = controller.getApiResponse();

		// Log the response.
		String resultCode = null;
		String messageCode = null;
		String resultMessage = null;
		if (response != null) {
			if (response.getMessages() != null) {
				resultCode = response.getMessages().getResultCode().value();
				MessageTypeEnum responseCode = response.getMessages().getResultCode();
				log.info("Update Payment Profile Result Code: " + responseCode);
				if (response.getMessages() != null && response.getMessages().getMessage() != null
						&& response.getMessages().getMessage().get(0) != null) {
					messageCode = response.getMessages().getMessage().get(0).getCode();
					resultMessage = response.getMessages().getMessage().get(0).getText();
					log.info("Update Payment Profile resultCode: " + resultCode);
					log.info("Update Payment Profile Message Code: " + messageCode);
					log.info("Update Payment Profile Message Text: " + resultMessage);
				}
			}
		}
		Map result = new HashMap();
		result.put("resultCode", resultCode);
		result.put("messageCode", messageCode);
		result.put("resultMessage", resultMessage);
		return result;
	}

	public Map deletePaymentProfile(Map parameters) {

		// Test or live mode
		Environment env = configureEnvironment(parameters);
		ApiOperationBase.setEnvironment(env);

		// Authentication
		ApiOperationBase.setMerchantAuthentication(configureMerchantAuthentication(parameters));

		// Create the request
		DeleteCustomerPaymentProfileRequest apiRequest = new DeleteCustomerPaymentProfileRequest();
		String customerProfileID = null;
		if (StringUtils.isNotBlank((String) parameters.get("customerProfileID"))) {
			customerProfileID = (String) parameters.get("customerProfileID");
		} else {
			customerProfileID = getUser().getCustomer().getCimProfileID();
		}
		apiRequest.setCustomerProfileId(customerProfileID);
		apiRequest.setMerchantAuthentication(configureMerchantAuthentication(parameters));
		apiRequest.setCustomerPaymentProfileId((String) parameters.get("paymentProfileID"));

		// Send the request
		DeleteCustomerPaymentProfileController controller = new DeleteCustomerPaymentProfileController(apiRequest);
		controller.execute();
		DeleteCustomerPaymentProfileResponse response = controller.getApiResponse();

		// Log the response.
		String resultCode = null;
		String messageCode = null;
		String resultMessage = null;
		if (response != null) {
			if (response.getMessages() != null) {
				resultCode = response.getMessages().getResultCode().value();
				MessageTypeEnum responseCode = response.getMessages().getResultCode();
				log.info("Delete Payment Profile Result Code: " + responseCode);
				if (response.getMessages() != null && response.getMessages().getMessage() != null
						&& response.getMessages().getMessage().get(0) != null) {
					messageCode = response.getMessages().getMessage().get(0).getCode();
					resultMessage = response.getMessages().getMessage().get(0).getText();
					log.info("Delete Payment Profile resultCode: " + resultCode);
					log.info("Delete Payment Profile Message Code: " + messageCode);
					log.info("Delete Payment Profile Message Text: " + resultMessage);
				}
			}
		}
		Map result = new HashMap();
		result.put("resultCode", resultCode);
		result.put("messageCode", messageCode);
		result.put("resultMessage", resultMessage);
		return result;
	}

	public Map hostedProfilePage(Map parameters) {

		// Test or live mode
		Environment env = configureEnvironment(parameters);
		ApiOperationBase.setEnvironment(env);

		// Authentication
		ApiOperationBase.setMerchantAuthentication(configureMerchantAuthentication(parameters));

		// Create the request
		GetHostedProfilePageRequest apiRequest = new GetHostedProfilePageRequest();
		apiRequest.setCustomerProfileId((String) parameters.get("cimProfileID"));

		// Settings
		ArrayOfSetting alist = new ArrayOfSetting();

		// authorizeNetHostedIFrameCommunicatorUrl
		String authorizeNetHostedIFrameCommunicatorUrl = getSettings()
				.getValue("authorizeNetHostedIFrameCommunicatorUrl");
		String customerSecureURL = getSettings().getValue("customerSecureURL");
		SettingType setting = new SettingType();
		setting.setSettingName("hostedProfileIFrameCommunicatorUrl");
		setting.setSettingValue(customerSecureURL + authorizeNetHostedIFrameCommunicatorUrl);
		alist.getSetting().add(setting);

		// hostedProfileBillingAddressRequired
		// Note: See below. The Street Address and ZIP are ALWAYS required regardless of this setting. This affects
		// whether the first name, last name, city, and state are required though.
		String hostedProfileBillingAddressRequired = getSettings().getValue("authorizeNetHostedRequireBillingAddress");
		setting = new SettingType();
		setting.setSettingName("hostedProfileBillingAddressRequired");
		setting.setSettingValue("0".equals(hostedProfileBillingAddressRequired) ? "false" : "true");
		alist.getSetting().add(setting);

		// The billing fields MUST be shown because Authorize.net will send back the "Address and Zip are required."
		// error.
		// This happens regardless of various settings. It appears to be an Authorize.net bug with multiple forum
		// threads complaining about it.
		// hostedProfileBillingAddressOptions
		// String hostedProfileShowBillingAddress = getSettings().getValue("authorizeNetHostedShowBillingAddress");
		// setting = new SettingType();
		// setting.setSettingName("hostedProfileBillingAddressOptions");
		// setting.setSettingValue("0".equals(hostedProfileShowBillingAddress) ? "showNone" : "showBillingAddress");
		// alist.getSetting().add(setting);

		apiRequest.setHostedProfileSettings(alist);

		// Send the request
		GetHostedProfilePageController controller = new GetHostedProfilePageController(apiRequest);
		controller.execute();
		GetHostedProfilePageResponse response = controller.getApiResponse();

		// Log the response.
		String resultCode = null;
		String messageCode = null;
		String resultMessage = null;
		String accountHostedPageToken = null;
		if (response != null) {
			accountHostedPageToken = response.getToken();
			if (response.getMessages() != null) {
				resultCode = response.getMessages().getResultCode().value();
				MessageTypeEnum responseCode = response.getMessages().getResultCode();
				log.info("GetHostedProfilePage Result Code: " + responseCode);
				if (response.getMessages() != null && response.getMessages().getMessage() != null
						&& response.getMessages().getMessage().get(0) != null) {
					messageCode = response.getMessages().getMessage().get(0).getCode();
					resultMessage = response.getMessages().getMessage().get(0).getText();
					log.info("GetHostedProfilePage resultCode: " + resultCode);
					log.info("GetHostedProfilePage Message Code: " + messageCode);
					log.info("GetHostedProfilePage Message Text: " + resultMessage);
				}
			}
		}
		Map result = new HashMap();
		result.put("resultCode", resultCode);
		result.put("messageCode", messageCode);
		result.put("resultMessage", resultMessage);
		result.put("accountHostedPageToken", accountHostedPageToken);
		return result;
	}

	public Map hostedPaymentPage(Map parameters) {

		// Test or live mode
		Environment env = configureEnvironment(parameters);
		ApiOperationBase.setEnvironment(env);

		// Authentication
		ApiOperationBase.setMerchantAuthentication(configureMerchantAuthentication(parameters));

		// Create the request
		GetHostedPaymentPageRequest apiRequest = new GetHostedPaymentPageRequest();

		// Transaction type from settings
		String authorizeNetTxType = (String) getSettings().getValue("authorizeNetTxType");
		if (authorizeNetTxType.equals("AUTH_CAPTURE")) {
			parameters.put("TransactionType", TransactionTypeEnum.AUTH_CAPTURE_TRANSACTION.value());
		} else {
			parameters.put("TransactionType", TransactionTypeEnum.AUTH_ONLY_TRANSACTION.value());
		}

		// Instantiate the transaction request
		TransactionRequestType txnRequest = initTransactionRequest(parameters);

		// Populate the solution ID
		txnRequest.setSolution(configureSolution(ApiOperationBase.getEnvironment()));

		// Pass in the customer's profile id if they have one
		if ("1".equals(getSettings().getValue("authorizeNetEnableCustomerProfiles"))) {
			CustomerProfilePaymentType profileType = new CustomerProfilePaymentType();
			String customerProfileID = null;
			if (StringUtils.isNotBlank((String) parameters.get("customerProfileID"))) {
				customerProfileID = (String) parameters.get("customerProfileID");
			} else {
				customerProfileID = getUser().getCustomer().getCimProfileID();
			}
			if (StringUtils.isNotBlank(customerProfileID)) {
				profileType.setCustomerProfileId(customerProfileID);
				txnRequest.setProfile(profileType);
			}
		}

		// Customer name and billing address fields
		if (getUser() != null && getUser().getOrder() != null) {
			txnRequest.setBillTo(configureBillingAddress(parameters));
		}

		// Total amount, tax and shipping
		configureTotals(txnRequest, parameters);

		// Invoice and customer reference fields
		configureReferenceFields(txnRequest, parameters);

		// Reset the customerId reference field to a random string and also put it in the user's session. This is used
		// to validate the response when it is returned from the iframe. The customerId is the only field we can use
		// because it's the only field we can adjust that's part of both the request and the response. RefId or
		// TransRefId would have been much better but they are not part of the response for some reason.
		configureCustomerIdInSessionAndTransactionRequest(txnRequest);

		// Transaction settings such as email control fields and the duplicate transaction window
		configureTransactionSettings(txnRequest, parameters);

		// Customer shipping fields
		if (getUser() != null && getUser().getOrder() != null && getUser().getOrder().firstDelivery() != null) {
			txnRequest.setShipTo(configureShippingAddress(parameters));
		}
		apiRequest.setTransactionRequest(txnRequest);

		// Settings
		ArrayOfSetting alist = new ArrayOfSetting();
		String customerSecureURL = getSettings().getValue("customerSecureURL");
		String authorizeNetHostedPayText = getSettings().getValue("authorizeNetHostedPayText");
		String authorizeNetIntegrationMethod = getSettings().getValue("authorizeNetIntegrationMethod");
		String authorizeNetHostedIFrameCommunicatorUrl = getSettings()
				.getValue("authorizeNetHostedIFrameCommunicatorUrl");
		String authorizeNetHostedBgColor = getSettings().getValue("authorizeNetHostedBgColor");
		String securityCodeRequired = getSettings().getValue("securityCodeRequired");
		String authorizeNetHostedShowCreditCard = getSettings().getValue("authorizeNetHostedShowCreditCard");
		String authorizeNetHostedShowBankAccount = getSettings().getValue("authorizeNetHostedShowBankAccount");
		String useImageVerificationWithCCInfo = getSettings().getValue("useImageVerificationWithCCInfo");
		String authorizeNetHostedShowBillingAddress = getSettings().getValue("authorizeNetHostedShowBillingAddress");
		String authorizeNetHostedRequireBillingAddress = getSettings()
				.getValue("authorizeNetHostedRequireBillingAddress");
		String authorizeNetHostedShowEmail = getSettings().getValue("authorizeNetHostedShowEmail");
		String authorizeNetHostedRequireEmail = getSettings().getValue("authorizeNetHostedRequireEmail");
		String authorizeNetHostedAddPaymentProfile = getSettings().getValue("authorizeNetHostedAddPaymentProfile");

		// hostedPaymentButtonOptions
		if (StringUtils.isNotBlank(authorizeNetHostedPayText)) {
			Map hostedPaymentButtonOptions = new HashMap();
			hostedPaymentButtonOptions.put("text", authorizeNetHostedPayText);
			SettingType hostedPaymentButton = createJsonSetting("hostedPaymentButtonOptions",
					hostedPaymentButtonOptions);
			alist.getSetting().add(hostedPaymentButton);
		}

		// hostedPaymentStyleOptions
		if (StringUtils.isNotBlank(authorizeNetHostedBgColor)) {
			Map hostedPaymentStyleOptions = new HashMap();
			hostedPaymentStyleOptions.put("bgColor", authorizeNetHostedBgColor);
			SettingType hostedPaymentStyle = createJsonSetting("hostedPaymentStyleOptions", hostedPaymentStyleOptions);
			alist.getSetting().add(hostedPaymentStyle);
		}

		// hostedPaymentPaymentOptions - use the card code setting from main Payment Settings screen
		Map hostedPaymentPaymentOptions = new HashMap();
		hostedPaymentPaymentOptions.put("cardCodeRequired", "0".equals(securityCodeRequired) ? false : true);
		hostedPaymentPaymentOptions.put("showCreditCard", "0".equals(authorizeNetHostedShowCreditCard) ? false : true);
		hostedPaymentPaymentOptions.put("showBankAccount",
				"0".equals(authorizeNetHostedShowBankAccount) ? false : true);
		SettingType hostedPaymentPayment = createJsonSetting("hostedPaymentPaymentOptions",
				hostedPaymentPaymentOptions);
		alist.getSetting().add(hostedPaymentPayment);

		// hostedPaymentSecurityOptions - use the captcha setting from Security Settings -> Captcha Verification
		if ("1".equals(useImageVerificationWithCCInfo)) {
			Map hostedPaymentSecurityOptions = new HashMap();
			hostedPaymentSecurityOptions.put("captcha", true);
			SettingType hostedPaymentSecurity = createJsonSetting("hostedPaymentSecurityOptions",
					hostedPaymentSecurityOptions);
			alist.getSetting().add(hostedPaymentSecurity);
		}

		// hostedPaymentShippingAddressOptions - use the defaults of {"show": false, "required": false}

		// hostedPaymentBillingAddressOptions
		Map hostedPaymentBillingAddressOptions = new HashMap();
		hostedPaymentBillingAddressOptions.put("show", "0".equals(authorizeNetHostedShowBillingAddress) ? false : true);
		hostedPaymentBillingAddressOptions.put("required",
				"0".equals(authorizeNetHostedRequireBillingAddress) ? false : true);
		SettingType hostedPaymentBillingAddress = createJsonSetting("hostedPaymentBillingAddressOptions",
				hostedPaymentBillingAddressOptions);
		alist.getSetting().add(hostedPaymentBillingAddress);

		// hostedPaymentCustomerOptions
		Map hostedPaymentCustomerOptions = new HashMap();
		hostedPaymentCustomerOptions.put("showEmail", "0".equals(authorizeNetHostedShowEmail) ? false : true);
		hostedPaymentCustomerOptions.put("requiredEmail", "0".equals(authorizeNetHostedRequireEmail) ? false : true);
		hostedPaymentCustomerOptions.put("addPaymentProfile",
				"0".equals(authorizeNetHostedAddPaymentProfile) ? false : true);
		SettingType hostedPaymentCustomer = createJsonSetting("hostedPaymentCustomerOptions",
				hostedPaymentCustomerOptions);
		alist.getSetting().add(hostedPaymentCustomer);

		if (authorizeNetIntegrationMethod.equals("AcceptHostedIframe")) {

			// hostedPaymentReturnOptions
			Map hostedPaymentReturnOptions = new HashMap();
			hostedPaymentReturnOptions.put("showReceipt", false);
			SettingType hostedPaymentReturn = createJsonSetting("hostedPaymentReturnOptions",
					hostedPaymentReturnOptions);
			alist.getSetting().add(hostedPaymentReturn);

			// hostedPaymentIFrameCommunicatorOptions
			Map hostedPaymentIFrameCommunicatorOptions = new HashMap();
			hostedPaymentIFrameCommunicatorOptions.put("url",
					customerSecureURL + authorizeNetHostedIFrameCommunicatorUrl);
			SettingType hostedPaymentIFrameCommunicatorUrl = createJsonSetting("hostedPaymentIFrameCommunicatorUrl",
					hostedPaymentIFrameCommunicatorOptions);
			alist.getSetting().add(hostedPaymentIFrameCommunicatorUrl);

		}
		apiRequest.setHostedPaymentSettings(alist);

		// Send the request
		GetHostedPaymentPageController controller = new GetHostedPaymentPageController(apiRequest);
		controller.execute();
		GetHostedPaymentPageResponse response = controller.getApiResponse();

		// Log the response.
		String resultCode = null;
		String messageCode = null;
		String resultMessage = null;
		String hostedPageToken = null;
		if (response != null) {
			hostedPageToken = response.getToken();
			if (response.getMessages() != null) {
				resultCode = response.getMessages().getResultCode().value();
				MessageTypeEnum responseCode = response.getMessages().getResultCode();
				log.info("GetHostedPaymentPage Result Code: " + responseCode);
				if (response.getMessages() != null && response.getMessages().getMessage() != null
						&& response.getMessages().getMessage().get(0) != null) {
					messageCode = response.getMessages().getMessage().get(0).getCode();
					resultMessage = response.getMessages().getMessage().get(0).getText();
					log.info("GetHostedPaymentPage resultCode: " + resultCode);
					log.info("GetHostedPaymentPage Message Code: " + messageCode);
					log.info("GetHostedPaymentPage Message Text: " + resultMessage);
				}
			}
		}
		Map result = new HashMap();
		result.put("resultCode", resultCode);
		result.put("messageCode", messageCode);
		result.put("resultMessage", resultMessage);
		result.put("hostedPageToken", hostedPageToken);
		return result;
	}

	private String configureCustomerIdInSessionAndTransactionRequest(TransactionRequestType txnRequest) {
		String refId = UUID.randomUUID().toString().substring(0, 20);
		if (getUser().getCustomMap() == null) {
			getUser().setCustomMap(new HashMap());
		}
		getUser().getCustomMap().put("AcceptHostedCustomerId", refId);
		CustomerDataType customer = new CustomerDataType();
		customer.setId(refId);
		txnRequest.setCustomer(customer);
		return refId;
	}

	private SettingType createJsonSetting(String settingName, Map options) {
		SettingType setting = new SettingType();
		try {
			setting.setSettingName(settingName);
			ObjectMapper mapper = new ObjectMapper();
			setting.setSettingValue(mapper.writeValueAsString(options));
		} catch (Exception e) {
			log.error("Exception creating JSON string (swallowing): ", e);
		}
		return setting;
	}
}
