package com.softslate.commerce.businessobjects.shipping;

import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;

import com.softslate.commerce.businessobjects.core.ApplicationException;
import com.softslate.commerce.businessobjects.order.OrderShippingRule;
import com.softslate.commerce.businessobjects.product.Manufacturer;

/**
 * Connects to the UPS online server for live shipping rates.
 * <p>
 * <code>UPSShippingProcessor</code> is an implementation of the
 * <code>ShippingProcessor</code> interface for the application.
 * 
 * @author Jason McWilliams, David Tobey
 * @see com.softslate.commerce.businessobjects.shipping.ShippingProcessor
 */
public class UPSShippingProcessor extends BaseShippingProcessor implements
		ShippingProcessor {

	static Log log = LogFactory.getLog(UPSShippingProcessor.class);

	/*
	 * The total weight of the shipment - all the package weights added
	 * together.
	 */
	public Double totalShipmentWeight = new Double(0);

	public String UPS_Url = null;

	/*
	 * UPS Settings for requests
	 */
	public String UPS_accessLicenseNumber = null;

	public String UPS_userId = null;

	public String UPS_password = null;

	public String UPS_requestAction = "Rate";

	/*
	 * Request Option Rate = The server rates the specific service type provided
	 * and validates the shipment. This is the default behavior if an option is
	 * not provided. Shop = The server validates the shipment, and return rates
	 * for all UPS services from the ShipFrom to the ShipTo addresses. If a
	 * ShipFrom address is not present, it is assumed that the shipment is
	 * shipped from the shippers address.
	 */
	public String UPS_requestOption = "Shop";

	/*
	 * Pickup Type Code. Default value is 01. Valid values are: 01- Daily Pickup
	 * 03- Customer Counter 06- One Time Pickup 07- On Call Air 11- Suggested
	 * Retail Rates 19- Letter Center 20- Air Service Center
	 */
	public String UPS_pickupTypeCode = null;

	/*
	 * Customer Classification Code Valid values are: 01 - Wholesale 03 -
	 * Occasional 04 - Retail
	 */
	public String UPS_customerClassificationCode = null;

	/*
	 * Shipper Name
	 */
	public String UPS_shipperName = null;

	/*
	 * Shipper Attention to
	 */
	public String UPS_shipperAttentionName = null;

	/*
	 * Tax ID Number
	 */
	public String UPS_shipperTaxIdNumber = null;

	/*
	 * Phone number
	 */
	public String UPS_shipperPhoneNumber = null;

	/*
	 * Fax Number
	 */
	public String UPS_shipperFaxNumber = null;

	/*
	 * Shipper Number The shippers UPS account number. When the Shipper account
	 * is a daily pickup account or an occasional account in US, the shipment
	 * will be evaluated for Hundredweight rating. A shipper number is required
	 * when requesting to receive the negotiated rates. The shipper number is
	 * optional when requesting to receive the published rates.
	 */
	public String UPS_shipperNumber = null;

	/*
	 * The shippers city. Required if country does not utilize postal codes
	 */
	public String UPS_shipperCity = null;

	/*
	 * The shippers state code. Required if the ShipFrom container is absent.
	 */
	public String UPS_shipperStateCode = null;

	/*
	 * The shippers postal code. Required if country utilizes postal codes (i.e.
	 * US PR)
	 */
	public String UPS_shipperPostalCode = null;

	/*
	 * The UPS billing code for the shippers country. See Origin Country
	 * Information table. Default = US.
	 */
	public String UPS_shipperCountryCode = null;

	/*
	 * A flag indicating if the receivers address is a residential location. For
	 * our uses, use the following 0 = No residential 1 = Only residential 2 =
	 * Use orginization setting on checkout
	 */
	public int UPS_residentialAddressIndicator = 2;

	/*
	 * The senders city. Must be supplied if the senders country does not have a
	 * Postal Code.
	 */
	public String UPS_shipFromCity = null;

	/*
	 * The senders state code. Required if the ShipFrom container is present.
	 */
	public String UPS_shipFromStateCode = null;

	/*
	 * The senders postal code. Required if Ship To country code is US. Must
	 * supply either the senders city or postal code.
	 */
	public String UPS_shipFromPostalCode = null;

	/*
	 * The UPS billing code representing the senders country. See Appendix C for
	 * UPS Country Code table. Default = US.
	 */
	public String UPS_shipFromCountryCode = null;

	/*
	 * The code representing the unit of measure associated with the shipment
	 * weight. Shipment weight unit of measurement code. Codes are: LBS =
	 * Pounds, KGS = Kilograms, 00 = Metric Unit of Measurements, 01 = English
	 * Unit of Measurements Defaults Unit of Measurement used in the shippers
	 * country
	 */
	public String UPS_unitOfMeasurementCode = null;

	/*
	 * The code for the UPS service associated with the shipment. All valid UPS
	 * Service Codes are available in Appendix C. Only looked at if
	 * RequestOption = rate not shop 01 UPS Next Day Air 02 UPS 2nd Day Air 03
	 * UPS Ground 07 UPS Worldwide Express 08 UPS Worldwide Expedited 11 UPS
	 * Standard 12 UPS 3 Day Select 13 UPS Next Day Air Saver 14 UPS Next Day
	 * Air Early A.M. 54 UPS Worldwide Express Plus 59 UPS 2nd Day Air AM
	 */
	public String UPS_validServiceCodes = null;

	// NOTE We could an array here because we can ship multiple packages.
	// For now, we are doing just one package
	/*
	 * The code for the UPS packaging type associated with the package. All
	 * valid UPS Packaging Types are available in Appendix C. 01 UPS letter /
	 * UPSExpressEnvelope 02 Package 03 UPS Tube 04 UPS Pak 21 UPS Express Box
	 * 24 UPS 25Kg Box 25 UPS 10Kg Box
	 */
	public String UPS_packageCodeType = null;

	public boolean UPS_packageUseDimensions = false;

	/*
	 * The code for the unit of measure associated with the package dimensions.
	 * IN = Inches (default) DOES NOT HAVE TO BE INCLUDED IF WE SKIP THE Package
	 * Dimensions NODE
	 */
	public String UPS_packageDimensionsUomCode = "IN";

	/*
	 * The length of the package used to determine dimensional weight. Valid
	 * values are between 1 and 108.00. DOES NOT HAVE TO BE INCLUDED IF WE SKIP
	 * THE Package Dimensions NODE
	 */
	public String UPS_packageDimensionsLength = "";

	/*
	 * The height of the package used to determine dimensional weight. Valid
	 * values are between 1 and 108.00. DOES NOT HAVE TO BE INCLUDED IF WE SKIP
	 * THE Package Dimensions NODE
	 */
	public String UPS_packageDimensionsHeight = "";

	/*
	 * The width of the package used to determine dimensional weight. Valid
	 * values are between 1 and 108.00. DOES NOT HAVE TO BE INCLUDED IF WE SKIP
	 * THE Package Dimensions NODE
	 */
	public String UPS_packageDimensionsWidth = "";

	/*
	 * Code representing package size. 1 = OS1, Oversize 1 (package exceeds 84
	 * inches and is equal to or less than 108 inches in length and girth
	 * combined and weighs less than 30 pounds) 2 = OS2, Oversize 2 (package
	 * exceeds 108 inches in length and girth combined and weighs less than 70
	 * pounds) 3 = OS3, Oversize 3. (package exceeds 130 inches in length and is
	 * equal to or less than 165 in length and girth combined and weighs less
	 * than 90 pounds.)
	 */
	public String UPS_packageOversizePackage = "";

	/*
	 * Negotiated Rates option indicator. If the indicator is present and the
	 * Shipper is qualified, the Rates tool will return Negotiated Rates in the
	 * response.
	 */
	public boolean UPS_rateInformationNegotiatedRatesIndicator = false;

	/*
	 * The IATA currency code associated with the declared value amount for the
	 * package. Required if a value for the package declared value amount exists
	 * in the MonetaryValue tag. Must match one of the IATA currency codes as
	 * listed in Appendix C.
	 */
	public String UPS_insuredValueCurrencyCode = null;

	/*
	 * Use the insured value tag? Max value of 50,000.00 USD.
	 */
	public boolean UPS_useInsuredValue = true;

	/*
	 * Extra charge added to the rates returned by UPS
	 */
	public String UPS_extraCharge = null;

	/*
	 * Extra weight added to the pacakages whose rates are requested
	 */
	public String UPS_extraWeight = null;

	/*
	 * If defined, package weights will be capped at this value
	 */
	public Double UPS_maxPackageWeight = null;

	/*
	 * UPS SHIPPING METHODS
	 */
	public Map UPS_MethodsMap = Collections.synchronizedMap(new HashMap());

	public Collection loadShippingOptions() throws Exception {
		if (log.isInfoEnabled())
			log.info("Loading shipping options.");
		setUPSSettings();
		setUpsMethodsMap();

		// Throw an error if a delivery address is not defined.
		if (getUser().getOrder() == null
				|| getUser().getOrder().firstDelivery() == null
				|| getUser().getOrder().firstDelivery().getCountry() == null
				|| getUser().getOrder().firstDelivery().getState() == null
				|| getUser().getOrder().firstDelivery().getPostalCode() == null) {
			if (log.isInfoEnabled())
				log.info("Delivery address not defined.");
			String errorCode = "ADDRESS";
			String errorMessage = "Delivery address not defined.";
			throw new ApplicationException(errorCode + ": " + errorMessage);
		}

		// A collection where all the rates from each drop shipper will be added
		// up
		Collection mergedRates = new ArrayList();

		// Find all the drop shippers represented by items in the order, and
		// define the dropShipperIDs variable.
		retrieveDropShippersFromOrder(getUser().getOrder());

		// First query for the shipping rates for everything that does not have
		// a drop shipper, using the global settings.
		if (log.isInfoEnabled())
			log.info("Processing non-drop shipped items.");
		currentDropShipperID = null;
		String methodsAndRates = getXmlFromUps();
		if (methodsAndRates != null) {
			Map ratesAndErrors = formatRates(methodsAndRates);
			Collection rates = (Collection) ratesAndErrors.get("rates");
			String errorCode = (String) ratesAndErrors.get("errorCode");
			String errorMessage = (String) ratesAndErrors.get("errorMessage");
			if ((rates == null || rates.isEmpty()) && errorCode != null
					&& errorMessage != null) {
				throw new ApplicationException(errorCode + ": " + errorMessage);
			}

			// Merge the rates for all drop shippers by adding them up
			mergedRates = mergeRates(mergedRates, rates);
		}

		// Query for each drop shipper now separately, adding the rates together
		// as we go.
		Iterator i = dropShippers.iterator();
		while (i.hasNext()) {
			Manufacturer m = (Manufacturer) i.next();
			UPS_shipFromCity = m.getCity();
			UPS_shipFromStateCode = m.getState();
			UPS_shipFromPostalCode = m.getPostalCode();
			UPS_shipFromCountryCode = m.getCountry();
			currentDropShipperID = new Integer(m.getManufacturerID());
			totalShipmentWeight = new Double(0);
			if (log.isInfoEnabled())
				log.info("Processing drop shipper: " + m.getCode());
			String dropShipperMethodsAndRates = getXmlFromUps();
			if (dropShipperMethodsAndRates != null) {
				Map dropShipperRatesAndErrors = formatRates(dropShipperMethodsAndRates);

				// Throw an exception if there was an error and no rates
				// returned
				Collection dropShipperRates = (Collection) dropShipperRatesAndErrors
						.get("rates");
				String dropShipperErrorCode = (String) dropShipperRatesAndErrors
						.get("errorCode");
				String dropShipperErrorMessage = (String) dropShipperRatesAndErrors
						.get("errorMessage");
				if ((dropShipperRates == null || dropShipperRates.isEmpty())
						&& dropShipperErrorCode != null
						&& dropShipperErrorMessage != null) {
					throw new ApplicationException(dropShipperErrorCode + ": "
							+ dropShipperErrorMessage);
				}

				// Merge the rates for all drop shippers by adding them up
				mergedRates = mergeRates(mergedRates, dropShipperRates);
			}

		}

		// Apply the extra charge as defined in the settings
		mergedRates = doGlobalExtraCharge(mergedRates);

		return mergedRates;
	}

	public void setUPSSettings() throws Exception {
		UPS_Url = (String) getSettings().getValue("UPS_Url");
		UPS_accessLicenseNumber = (String) getSettings().getValue(
				"UPS_accessLicenseNumber");
		UPS_userId = (String) getSettings().getValue("UPS_userId");
		UPS_password = (String) getSettings().getValue("UPS_password");
		UPS_pickupTypeCode = (String) getSettings().getValue(
				"UPS_pickupTypeCode");
		UPS_customerClassificationCode = (String) getSettings().getValue(
				"UPS_customerClassificationCode");
		UPS_shipperName = (String) getSettings().getValue("UPS_shipperName");
		UPS_shipperAttentionName = (String) getSettings().getValue(
				"UPS_shipperAttentionName");
		UPS_shipperTaxIdNumber = (String) getSettings().getValue(
				"UPS_shipperTaxIdNumber");
		UPS_shipperPhoneNumber = (String) getSettings().getValue(
				"UPS_shipperPhoneNumber");
		UPS_shipperFaxNumber = (String) getSettings().getValue(
				"UPS_shipperFaxNumber");
		UPS_shipperNumber = (String) getSettings()
				.getValue("UPS_shipperNumber");
		UPS_shipperCity = (String) getSettings().getValue("UPS_shipperCity");
		UPS_shipperStateCode = (String) getSettings().getValue(
				"UPS_shipperStateCode");
		UPS_shipperPostalCode = (String) getSettings().getValue(
				"UPS_shipperPostalCode");
		UPS_shipperCountryCode = (String) getSettings().getValue(
				"UPS_shipperCountryCode");
		UPS_residentialAddressIndicator = Integer.parseInt(getSettings()
				.getValue("UPS_residentialAddressIndicator").toString());
		UPS_shipFromCity = (String) getSettings().getValue("UPS_shipFromCity");
		UPS_shipFromStateCode = (String) getSettings().getValue(
				"UPS_shipFromStateCode");
		UPS_shipFromPostalCode = (String) getSettings().getValue(
				"UPS_shipFromPostalCode");
		UPS_shipFromCountryCode = (String) getSettings().getValue(
				"UPS_shipFromCountryCode");
		UPS_unitOfMeasurementCode = (String) getSettings().getValue(
				"UPS_unitOfMeasurementCode");
		UPS_packageCodeType = (String) getSettings().getValue(
				"UPS_packageCodeType");

		if (((String) getSettings().getValue("UPS_useInsuredValue"))
				.equals("Yes")) {
			UPS_useInsuredValue = true;
		} else {
			UPS_useInsuredValue = false;
		}
		UPS_insuredValueCurrencyCode = (String) getSettings().getValue(
				"UPS_insuredValueCurrencyCode");

		if (((String) getSettings().getValue(
				"UPS_rateInformationNegotiatedRatesIndicator")).equals("Yes")) {
			UPS_rateInformationNegotiatedRatesIndicator = true;
		} else {
			UPS_rateInformationNegotiatedRatesIndicator = false;
		}
		if (((String) getSettings().getValue("UPS_packageUseDimensions"))
				.equals("Yes")) {
			UPS_packageUseDimensions = true;
		} else {
			UPS_packageUseDimensions = false;
		}

		UPS_validServiceCodes = (String) getSettings().getValue(
				"UPS_validServiceCodes");

		UPS_packageDimensionsUomCode = (String) getSettings().getValue(
				"UPS_packageDimensionsUomCode");
		UPS_packageDimensionsLength = (String) getSettings().getValue(
				"UPS_packageDimensionsLength");
		UPS_packageDimensionsHeight = (String) getSettings().getValue(
				"UPS_packageDimensionsHeight");
		UPS_packageDimensionsWidth = (String) getSettings().getValue(
				"UPS_packageDimensionsWidth");
		UPS_extraCharge = (String) getSettings().getValue("USP_extraCharge");
		UPS_extraWeight = (String) getSettings().getValue("UPS_extraWeight");
		UPS_maxPackageWeight = new Double(150);
		try {
			UPS_maxPackageWeight = new Double((String) getSettings().getValue(
					"UPS_maxPackageWeight"));
		} catch (Exception e) {
			log.info("Error setting UPS max package weight: ", e);
		}

	}

	public void setUpsMethodsMap() throws Exception {
		/*
		 * Add the ups methods you want to allow here, along with names 01 UPS
		 * Next Day Air 02 UPS 2nd Day Air 03 UPS Ground 07 UPS Worldwide
		 * Express 08 UPS Worldwide Expedited 11 UPS Standard 12 UPS 3 Day
		 * Select 13 UPS Next Day Air Saver 14 UPS Next Day Air Early A.M. 54
		 * UPS Worldwide Express Plus 59 UPS 2nd Day Air AM
		 */
		UPS_MethodsMap.put("01", "UPS Next Day Air");
		UPS_MethodsMap.put("02", "UPS 2nd Day Air");
		UPS_MethodsMap.put("03", "UPS Ground");
		UPS_MethodsMap.put("07", "UPS Worldwide Express");
		UPS_MethodsMap.put("08", "UPS Worldwide Expedited");
		UPS_MethodsMap.put("11", "UPS Standard");
		UPS_MethodsMap.put("12", "UPS 3 Day Select");
		UPS_MethodsMap.put("13", "UPS Next Day Air Saver");
		UPS_MethodsMap.put("14", "UPS Next Day Air Early A.M.");
		UPS_MethodsMap.put("54", "UPS Worldwide Express Plus");
		UPS_MethodsMap.put("59", "UPS 2nd Day Air AM");
		UPS_MethodsMap.put("65", "UPS Worldwide Saver");

	}

	public int processShipping() throws Exception {
		return 0;
	}

	public String getXmlFromUps() throws Exception {
		if (log.isInfoEnabled())
			log.info("Loading active methods and rates from UPS.");

		// Get POST data from input StringBuffer containing an XML document
		String queryString = createUPSCall();
		if (queryString == null) {
			return null;
		}
		if (log.isDebugEnabled())
			log.debug("Request string: " + queryString);

		String data = sendHTTPPost(UPS_Url, queryString, true);
		StringBuffer xmlOut = new StringBuffer(data);
		return xmlOut.toString();

	}

	public String createUPSCall() throws Exception {

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		DOMImplementation impl = builder.getDOMImplementation();

		// Create access request container information
		Document document1 = createAccessRequest(impl);

		// Create the main request document
		Document document2 = createRatingServiceSelectionRequest(impl);
		if (document2 == null) {
			return null;
		}

		StringWriter output1 = new StringWriter();
		StringWriter output2 = new StringWriter();

		try {
			TransformerFactory.newInstance().newTransformer().transform(
					new DOMSource(document1), new StreamResult(output1));
			TransformerFactory.newInstance().newTransformer().transform(
					new DOMSource(document2), new StreamResult(output2));
		} catch (Exception ex) {
			log.error("Error transforming XML", ex);
		}
		return (output1.toString() + output2.toString());

	}

	public Document createAccessRequest(DOMImplementation impl) {

		// Add access request container information
		Document document1 = impl.createDocument(null, "AccessRequest", null);
		Element ar = document1.getDocumentElement();

		Element ar_AccessLicenseNumber = (Element) document1
				.createElement("AccessLicenseNumber");
		ar.appendChild(ar_AccessLicenseNumber);
		ar_AccessLicenseNumber.appendChild(document1
				.createTextNode(UPS_accessLicenseNumber));

		Element ar_UserId = (Element) document1.createElement("UserId");
		ar.appendChild(ar_UserId);
		ar_UserId.appendChild(document1.createTextNode(UPS_userId));

		Element ar_Password = (Element) document1.createElement("Password");
		ar.appendChild(ar_Password);
		ar_Password.appendChild(document1.createTextNode(UPS_password));

		return document1;
	}

	public Document createRatingServiceSelectionRequest(DOMImplementation impl) {

		boolean foundAPackage = false;

		Document document2 = impl.createDocument(null,
				"RatingServiceSelectionRequest", null);
		Element rs = document2.getDocumentElement();

		Element rs_Request = (Element) document2.createElement("Request");
		rs.appendChild(rs_Request);

		Element rs_RequestRequestAction = (Element) document2
				.createElement("RequestAction");
		rs_Request.appendChild(rs_RequestRequestAction);
		rs_RequestRequestAction.appendChild(document2
				.createTextNode(UPS_requestAction));

		Element rs_RequestRequestOption = (Element) document2
				.createElement("RequestOption");
		rs_Request.appendChild(rs_RequestRequestOption);
		rs_RequestRequestOption.appendChild(document2
				.createTextNode(UPS_requestOption));

		// Add pickup type container
		Element rs_PickupType = (Element) document2.createElement("PickupType");
		rs.appendChild(rs_PickupType);

		Element rs_PickupTypeCode = (Element) document2.createElement("Code");
		rs_PickupType.appendChild(rs_PickupTypeCode);
		rs_PickupTypeCode.appendChild(document2
				.createTextNode(UPS_pickupTypeCode));

		// add custom classification
		Element rs_CustomerClassification = (Element) document2
				.createElement("CustomerClassification");
		rs.appendChild(rs_CustomerClassification);

		Element rs_CustomerClassificationCode = (Element) document2
				.createElement("Code");
		rs_CustomerClassification.appendChild(rs_CustomerClassificationCode);
		rs_CustomerClassificationCode.appendChild(document2
				.createTextNode(UPS_customerClassificationCode));

		// Add the shipment
		Element rs_Shipment = (Element) document2.createElement("Shipment");
		rs.appendChild(rs_Shipment);

		// Add shipment/shipper
		Element rs_ShipmentShipper = createShipmentShipperNode(document2);
		rs_Shipment.appendChild(rs_ShipmentShipper);

		// Add Shipment/ShipTo
		Element rs_ShipmentShipTo = createShipmentShipToNode(document2);
		rs_Shipment.appendChild(rs_ShipmentShipTo);

		// Add Shipment/ShipFrom
		Element rs_ShipmentShipFrom = createShipmentShipFromNode(document2);
		rs_Shipment.appendChild(rs_ShipmentShipFrom);

		// Shipment Weight - leave it empty for now, fill it in below
		Element rs_ShipmentShipmentWeight = (Element) document2
				.createElement("ShipmentWeight");
		Element rs_ShipmentShipmentWeightUom = (Element) document2
				.createElement("UnitOfMeasurement");
		rs_ShipmentShipmentWeight.appendChild(rs_ShipmentShipmentWeightUom);
		Element rs_ShipmentShipmentWeightUomCode = (Element) document2
				.createElement("Code");
		rs_ShipmentShipmentWeightUom
				.appendChild(rs_ShipmentShipmentWeightUomCode);
		rs_ShipmentShipmentWeightUomCode.appendChild(document2
				.createTextNode(UPS_unitOfMeasurementCode));
		Element rs_ShipmentShipmentWeightWeight = (Element) document2
				.createElement("Weight");
		rs_ShipmentShipmentWeight.appendChild(rs_ShipmentShipmentWeightWeight);
		rs_Shipment.appendChild(rs_ShipmentShipmentWeight);

		// Package
		// Now add multiple packages from the orderShippingRule table
		if (getUser().getOrder().getOrderDeliveryCollection() != null
				&& !getUser().getOrder().getOrderDeliveryCollection().isEmpty()) {
			Iterator j = getUser().getOrder().getOrderShippingRules()
					.iterator();
			while (j.hasNext()) {
				OrderShippingRule osr = (OrderShippingRule) j.next();
				boolean createThePackage = isPackageOkToCreate(osr);
				if (createThePackage) {
					if (log.isInfoEnabled())
						log.info("Processing package: " + osr.getCode());
					Element rs_ShipmentPackage = createShipmentPackageNode(
							document2, osr);
					if (rs_ShipmentPackage != null) {
						rs_Shipment.appendChild(rs_ShipmentPackage);
						foundAPackage = true;
					}
				}
			}
		}

		// If no packages were found, return null to skip the call to UPS
		if (!foundAPackage) {
			return null;
		}

		// Now attach the total weight of the shipment to the empty node defined
		// above
		if (log.isDebugEnabled())
			log.debug("totalShipmentWeight: " + totalShipmentWeight);
		rs_ShipmentShipmentWeightWeight.appendChild(document2
				.createTextNode(totalShipmentWeight.toString()));

		// Negotiated Rates?
		if (UPS_rateInformationNegotiatedRatesIndicator) {
			Element rs_ShipmentRateInformation = (Element) document2
					.createElement("RateInformation");
			rs_Shipment.appendChild(rs_ShipmentRateInformation);

			Element rs_ShipmentRateInformationNri = (Element) document2
					.createElement("NegotiatedRatesIndicator");
			rs_ShipmentRateInformation
					.appendChild(rs_ShipmentRateInformationNri);
		}

		return document2;
	}

	public Element createShipmentShipperNode(Document document2) {
		Element rs_ShipmentShipper = (Element) document2
				.createElement("Shipper");

		Element rs_ShipmentShipperNumber = (Element) document2
				.createElement("ShipperNumber");
		rs_ShipmentShipper.appendChild(rs_ShipmentShipperNumber);
		rs_ShipmentShipperNumber.appendChild(document2
				.createTextNode(UPS_shipperNumber));

		Element rs_ShipmentShipperName = (Element) document2
				.createElement("Name");
		rs_ShipmentShipper.appendChild(rs_ShipmentShipperName);
		rs_ShipmentShipperName.appendChild(document2
				.createTextNode(UPS_shipperName));

		if (UPS_shipperAttentionName.length() > 0) {
			Element rs_ShipmentShipperAttentionName = (Element) document2
					.createElement("AttentionName");
			rs_ShipmentShipper.appendChild(rs_ShipmentShipperAttentionName);
			rs_ShipmentShipperAttentionName.appendChild(document2
					.createTextNode(UPS_shipperAttentionName));
		}

		Element rs_ShipmentShipperTaxIdNumber = (Element) document2
				.createElement("TaxIdentificationNumber");
		rs_ShipmentShipper.appendChild(rs_ShipmentShipperTaxIdNumber);
		rs_ShipmentShipperTaxIdNumber.appendChild(document2
				.createTextNode(UPS_shipperTaxIdNumber));

		Element rs_ShipmentShipperPhoneNumber = (Element) document2
				.createElement("PhoneNumber");
		rs_ShipmentShipper.appendChild(rs_ShipmentShipperPhoneNumber);
		rs_ShipmentShipperPhoneNumber.appendChild(document2
				.createTextNode(UPS_shipperPhoneNumber));

		Element rs_ShipmentShipperFaxNumber = (Element) document2
				.createElement("FaxNumber");
		rs_ShipmentShipper.appendChild(rs_ShipmentShipperFaxNumber);
		rs_ShipmentShipperFaxNumber.appendChild(document2
				.createTextNode(UPS_shipperFaxNumber));

		// Add Shipment/Shipper/Address
		Element rs_ShipmentShipperAddress = (Element) document2
				.createElement("Address");
		rs_ShipmentShipper.appendChild(rs_ShipmentShipperAddress);

		Element rs_ShipmentShipperAddressCity = (Element) document2
				.createElement("City");
		rs_ShipmentShipperAddress.appendChild(rs_ShipmentShipperAddressCity);
		rs_ShipmentShipperAddressCity.appendChild(document2
				.createTextNode(UPS_shipperCity));

		Element rs_ShipmentShipperAddressState = (Element) document2
				.createElement("StateProvinceCode");
		rs_ShipmentShipperAddress.appendChild(rs_ShipmentShipperAddressState);
		rs_ShipmentShipperAddressState.appendChild(document2
				.createTextNode(UPS_shipperStateCode));

		Element rs_ShipmentShipperAddressPostalCode = (Element) document2
				.createElement("PostalCode");
		rs_ShipmentShipperAddress
				.appendChild(rs_ShipmentShipperAddressPostalCode);
		rs_ShipmentShipperAddressPostalCode.appendChild(document2
				.createTextNode(UPS_shipperPostalCode));

		Element rs_ShipmentShipperAddressCountryCode = (Element) document2
				.createElement("CountryCode");
		rs_ShipmentShipperAddress
				.appendChild(rs_ShipmentShipperAddressCountryCode);
		rs_ShipmentShipperAddressCountryCode.appendChild(document2
				.createTextNode(UPS_shipperCountryCode));

		return rs_ShipmentShipper;
	}

	public Element createShipmentShipToNode(Document document2) {
		Element rs_ShipmentShipTo = (Element) document2.createElement("ShipTo");

		// Add Shipment/ShipTo/Address
		Element rs_ShipmentShipToAddress = (Element) document2
				.createElement("Address");
		rs_ShipmentShipTo.appendChild(rs_ShipmentShipToAddress);

		if (log.isDebugEnabled())
			log.debug("Ship to city: "
					+ getUser().getOrder().firstDelivery().getDeliveryCity());

		Element rs_ShipmentShipToAddressCity = (Element) document2
				.createElement("City");
		rs_ShipmentShipToAddress.appendChild(rs_ShipmentShipToAddressCity);
		rs_ShipmentShipToAddressCity.appendChild(document2
				.createTextNode(getUser().getOrder().firstDelivery()
						.getDeliveryCity()));

		Element rs_ShipmentShipToAddressStateProvinceCode = (Element) document2
				.createElement("StateProvinceCode");
		rs_ShipmentShipToAddress
				.appendChild(rs_ShipmentShipToAddressStateProvinceCode);
		rs_ShipmentShipToAddressStateProvinceCode.appendChild(document2
				.createTextNode(getUser().getOrder().firstDelivery()
						.getDeliveryState()));

		Element rs_ShipmentShipToAddressPostalCode = (Element) document2
				.createElement("PostalCode");
		rs_ShipmentShipToAddress
				.appendChild(rs_ShipmentShipToAddressPostalCode);
		rs_ShipmentShipToAddressPostalCode.appendChild(document2
				.createTextNode(getUser().getOrder().firstDelivery()
						.getDeliveryPostalCode()));

		Element rs_ShipmentShipToAddressCountryCode = (Element) document2
				.createElement("CountryCode");
		rs_ShipmentShipToAddress
				.appendChild(rs_ShipmentShipToAddressCountryCode);
		rs_ShipmentShipToAddressCountryCode.appendChild(document2
				.createTextNode(getUser().getOrder().firstDelivery()
						.getDeliveryCountry()));

		if (UPS_residentialAddressIndicator == 0) {
			// We dont do anything, we assume all shipments are to businesses
		}
		if (UPS_residentialAddressIndicator == 1) {
			// All shipments are residential
			Element rs_ShipmentShipToAddressResidential = (Element) document2
					.createElement("ResidentialAddressIndicator");
			rs_ShipmentShipToAddress
					.appendChild(rs_ShipmentShipToAddressResidential);
		}
		if (UPS_residentialAddressIndicator == 2) {
			// Check organization shipping info during checkout
			if (getUser().getOrder().firstDelivery().getOrganization() == null
					|| getUser().getOrder().firstDelivery().getOrganization().trim()
							.length() == 0) {
				Element rs_ShipmentShipToAddressResidential = (Element) document2
						.createElement("ResidentialAddressIndicator");
				rs_ShipmentShipToAddress
						.appendChild(rs_ShipmentShipToAddressResidential);
			}

		}

		return rs_ShipmentShipTo;
	}

	public Element createShipmentShipFromNode(Document document2) {

		// Add Shipment/ShipFrom/Address
		if (log.isDebugEnabled())
			log.debug("Ship from city: " + UPS_shipFromCity);

		Element rs_ShipmentShipFrom = (Element) document2
				.createElement("ShipFrom");

		Element rs_ShipmentShipFromAddress = (Element) document2
				.createElement("Address");
		rs_ShipmentShipFrom.appendChild(rs_ShipmentShipFromAddress);

		Element rs_ShipmentShipFromAddressCity = (Element) document2
				.createElement("City");
		rs_ShipmentShipFromAddress.appendChild(rs_ShipmentShipFromAddressCity);
		rs_ShipmentShipFromAddressCity.appendChild(document2
				.createTextNode(UPS_shipFromCity));

		Element rs_ShipmentShipFromAddressStateProvinceCode = (Element) document2
				.createElement("StateProvinceCode");
		rs_ShipmentShipFromAddress
				.appendChild(rs_ShipmentShipFromAddressStateProvinceCode);
		rs_ShipmentShipFromAddressStateProvinceCode.appendChild(document2
				.createTextNode(UPS_shipFromStateCode));

		Element rs_ShipmentShipFromAddressPostalCode = (Element) document2
				.createElement("PostalCode");
		rs_ShipmentShipFromAddress
				.appendChild(rs_ShipmentShipFromAddressPostalCode);
		rs_ShipmentShipFromAddressPostalCode.appendChild(document2
				.createTextNode(UPS_shipFromPostalCode));

		Element rs_ShipmentShipFromAddressCountryCode = (Element) document2
				.createElement("CountryCode");
		rs_ShipmentShipFromAddress
				.appendChild(rs_ShipmentShipFromAddressCountryCode);
		rs_ShipmentShipFromAddressCountryCode.appendChild(document2
				.createTextNode(UPS_shipFromCountryCode));

		return rs_ShipmentShipFrom;
	}

	public Element createShipmentPackageNode(Document document2,
			OrderShippingRule osr) {

		Element rs_ShipmentPackage = (Element) document2
				.createElement("Package");

		Element rs_ShipmentPackageType = (Element) document2
				.createElement("PackagingType");
		rs_ShipmentPackage.appendChild(rs_ShipmentPackageType);

		// Package type (Code)
		// Use the pkgExtra1 field if it exists, otherwise default
		Element rs_ShipmentPackageTypeCode = (Element) document2
				.createElement("Code");
		rs_ShipmentPackageType.appendChild(rs_ShipmentPackageTypeCode);
		if (osr.getPkgExtra1() != null && osr.getPkgExtra1().length() > 0) {
			rs_ShipmentPackageTypeCode.appendChild(document2.createTextNode(osr
					.getPkgExtra1()));
		} else {
			rs_ShipmentPackageTypeCode.appendChild(document2
					.createTextNode(UPS_packageCodeType));
		}

		// Add Dimensions if necessary
		Element rs_ShipmentPackageDimensions = createShipmentPackageDimensionsNode(
				document2, osr);
		if (rs_ShipmentPackageDimensions != null) {
			rs_ShipmentPackage.appendChild(rs_ShipmentPackageDimensions);
		}

		// Add the package weight
		Element rs_ShipmentPackagePackageWeight = createShipmentPackageWeightNode(
				document2, osr);

		// If the package weight node is null, it means we are on the generic
		// package and no items matched the current drop shipper, so we skip
		// this package.
		if (rs_ShipmentPackagePackageWeight == null) {
			if (log.isDebugEnabled())
				log.debug("Generic package and no items match the dropshipper");
			return null;
		}
		rs_ShipmentPackage.appendChild(rs_ShipmentPackagePackageWeight);

		// We can use the oversize package flag in a rule by
		// Setting pkgExtra3 to 1,2,3
		String oversizedValue = "";
		if (osr.getPkgExtra3() != null && osr.getPkgExtra3().length() > 0) {
			if (log.isDebugEnabled())
				log.debug("Override oversize:" + osr.getPkgExtra3());
			oversizedValue = osr.getPkgExtra3();
		} else if (UPS_packageOversizePackage.length() > 0) {
			oversizedValue = UPS_packageOversizePackage;
		} else {
			oversizedValue = "";
		}

		if (oversizedValue.length() > 0) {
			if (log.isDebugEnabled())
				log.debug("Package is oversized.");
			Element rs_ShipmentPackageOversize = (Element) document2
					.createElement("OversizePackage");
			rs_ShipmentPackage.appendChild(rs_ShipmentPackageOversize);
			rs_ShipmentPackageOversize.appendChild(document2
					.createTextNode(oversizedValue));
		}

		// Add package service options
		Element rs_ShipmentPackagePso = (Element) document2
				.createElement("PackageServiceOptions");
		rs_ShipmentPackage.appendChild(rs_ShipmentPackagePso);

		// We can override useInsuredValue with getPkgExtra4
		boolean useInsuredValue = false;
		if (osr.getPkgExtra4() != null && osr.getPkgExtra4().length() > 0
				&& osr.getPkgExtra4().equals("1")) {
			useInsuredValue = true;
		} else if (osr.getPkgExtra4() != null
				&& osr.getPkgExtra4().length() > 0
				&& osr.getPkgExtra4().equals("0")) {
			useInsuredValue = false;
		} else if (UPS_useInsuredValue) {
			useInsuredValue = true;
		} else {
			useInsuredValue = false;
		}

		if (useInsuredValue) {

			String declaredValue = "10.00";
			// if (getUser().getOrder().getSubtotal() != null) {
			// declaredValue =
			// getUser().getOrder().getSubtotal().toString();
			// }
			if (osr.getSubtotal() != null) {
				declaredValue = osr.getSubtotal().toString();
			}
			if (declaredValue.charAt(declaredValue.length() - 2) == '.') {
				declaredValue = declaredValue + "0";
			}

			Element rs_ShipmentPackagePsoInsuredValue = (Element) document2
					.createElement("InsuredValue");
			rs_ShipmentPackagePso
					.appendChild(rs_ShipmentPackagePsoInsuredValue);

			Element rs_ShipmentPackagePsoInsuredValueCurrencyCode = (Element) document2
					.createElement("CurrencyCode");
			rs_ShipmentPackagePsoInsuredValue
					.appendChild(rs_ShipmentPackagePsoInsuredValueCurrencyCode);
			rs_ShipmentPackagePsoInsuredValueCurrencyCode.appendChild(document2
					.createTextNode(UPS_insuredValueCurrencyCode));

			Element rs_ShipmentPackagePsoInsuredValueMonetaryValue = (Element) document2
					.createElement("MonetaryValue");
			rs_ShipmentPackagePsoInsuredValue
					.appendChild(rs_ShipmentPackagePsoInsuredValueMonetaryValue);
			rs_ShipmentPackagePsoInsuredValueMonetaryValue
					.appendChild(document2.createTextNode(declaredValue));

		}
		return rs_ShipmentPackage;
	}

	public Element createShipmentPackageDimensionsNode(Document document2,
			OrderShippingRule osr) {

		// In this case, if it is a [generic] type, we will use the
		// generic settings, otherwise,
		// we will use what comes from the table. The
		// UPS_packageUseDimensions can be overridden
		// in the pkgExtra2 field.
		boolean doDimensions = false;
		if (osr.getPkgExtra2() != null && osr.getPkgExtra2().length() > 0
				&& osr.getPkgExtra2().equals("1")) {
			doDimensions = true;
		} else if (osr.getPkgExtra2() != null
				&& osr.getPkgExtra2().length() > 0
				&& osr.getPkgExtra2().equals("0")) {
			doDimensions = false;
		} else if (UPS_packageUseDimensions) {
			doDimensions = true;
		} else {
			doDimensions = false;
		}

		if (doDimensions) {
			if (log.isDebugEnabled())
				log.debug("Doing Demensions");
			Element rs_ShipmentPackageDimensions = (Element) document2
					.createElement("Dimensions");
			Element rs_ShipmentPackageDimensionsUom = (Element) document2
					.createElement("UnitOfMeasurement");
			rs_ShipmentPackageDimensions
					.appendChild(rs_ShipmentPackageDimensionsUom);

			Element rs_ShipmentPackageDimensionsUomCode = (Element) document2
					.createElement("Code");
			rs_ShipmentPackageDimensionsUom
					.appendChild(rs_ShipmentPackageDimensionsUomCode);
			rs_ShipmentPackageDimensionsUomCode.appendChild(document2
					.createTextNode(UPS_packageDimensionsUomCode));

			// For generic, we have to use the default package
			// settings
			if (osr.getCode().equals("[generic]")) {
				Element rs_ShipmentPackageDimensionsLength = (Element) document2
						.createElement("Length");
				rs_ShipmentPackageDimensions
						.appendChild(rs_ShipmentPackageDimensionsLength);
				rs_ShipmentPackageDimensionsLength.appendChild(document2
						.createTextNode(UPS_packageDimensionsLength));

				Element rs_ShipmentPackageDimensionsWidth = (Element) document2
						.createElement("Width");
				rs_ShipmentPackageDimensions
						.appendChild(rs_ShipmentPackageDimensionsWidth);
				rs_ShipmentPackageDimensionsWidth.appendChild(document2
						.createTextNode(UPS_packageDimensionsWidth));

				Element rs_ShipmentPackageDimensionsHeight = (Element) document2
						.createElement("Height");
				rs_ShipmentPackageDimensions
						.appendChild(rs_ShipmentPackageDimensionsHeight);
				rs_ShipmentPackageDimensionsHeight.appendChild(document2
						.createTextNode(UPS_packageDimensionsHeight));
			} else {
				Element rs_ShipmentPackageDimensionsLength = (Element) document2
						.createElement("Length");
				rs_ShipmentPackageDimensions
						.appendChild(rs_ShipmentPackageDimensionsLength);
				rs_ShipmentPackageDimensionsLength.appendChild(document2
						.createTextNode(osr.getPkgLength().toString()));

				Element rs_ShipmentPackageDimensionsWidth = (Element) document2
						.createElement("Width");
				rs_ShipmentPackageDimensions
						.appendChild(rs_ShipmentPackageDimensionsWidth);
				rs_ShipmentPackageDimensionsWidth.appendChild(document2
						.createTextNode(osr.getPkgWidth().toString()));

				Element rs_ShipmentPackageDimensionsHeight = (Element) document2
						.createElement("Height");
				rs_ShipmentPackageDimensions
						.appendChild(rs_ShipmentPackageDimensionsHeight);
				rs_ShipmentPackageDimensionsHeight.appendChild(document2
						.createTextNode(osr.getPkgHeight().toString()));
			}

			return rs_ShipmentPackageDimensions;
		}

		return null;
	}

	public Element createShipmentPackageWeightNode(Document document2,
			OrderShippingRule osr) {

		Element rs_ShipmentPackagePackageWeight = (Element) document2
				.createElement("PackageWeight");
		Element rs_ShipmentPackagePackageWeightUom = (Element) document2
				.createElement("UnitOfMeasurement");
		rs_ShipmentPackagePackageWeight
				.appendChild(rs_ShipmentPackagePackageWeightUom);

		Element rs_ShipmentPackagePackageWeightUomCode = (Element) document2
				.createElement("Code");
		rs_ShipmentPackagePackageWeightUom
				.appendChild(rs_ShipmentPackagePackageWeightUomCode);
		rs_ShipmentPackagePackageWeightUomCode.appendChild(document2
				.createTextNode(UPS_unitOfMeasurementCode));

		Double weightDouble = findPackageWeight(osr);

		// Package might have no items, so just return (skips creating the
		// package)
		if (weightDouble == null) {
			return null;
		}

		// Apply the extra weight global setting
		double weight = weightDouble.doubleValue();
		if (UPS_extraWeight != null) {
			weight = weight + Double.parseDouble(UPS_extraWeight);
		}

		// If weight is still zero, let's skip the package
		if (weight == 0) {
			return null;
		}

		// If weight is more than zero, but less than .1 pounds, let's push it
		// up to .1 pounds because that is UPS's minimum
		if (weight < .1) {
			weight = .1;
		}

		// Check the weight against the maximum allowed by UPS (150 pounds by
		// default)
		if (UPS_maxPackageWeight != null
				&& weight > UPS_maxPackageWeight.doubleValue()) {
			weight = UPS_maxPackageWeight.doubleValue();
		}

		// Create the node
		Element rs_ShipmentPackagePackageWeightWeight = (Element) document2
				.createElement("Weight");
		rs_ShipmentPackagePackageWeight
				.appendChild(rs_ShipmentPackagePackageWeightWeight);
		rs_ShipmentPackagePackageWeightWeight.appendChild(document2
				.createTextNode(Double.toString(weight)));

		// Keep track of the weight of the shipment as a whole
		totalShipmentWeight = new Double(totalShipmentWeight.doubleValue()
				+ weight);

		return rs_ShipmentPackagePackageWeight;
	}

	public Map formatRates(String xmlString) throws Exception {
		// Create a Map to return rates and errors
		Map ratesAndErrors = new HashMap();
		String errorCode = null;
		String errorMessage = null;
		Collection rates = new ArrayList();

		Map responseStatus = getResponseStatusFromXml(xmlString);
		if (Integer.parseInt(responseStatus.get("statusCode").toString()) == 0) {
			// Error
			errorCode = (String) responseStatus.get("errorCode");
			ratesAndErrors.put("rates", rates);
			ratesAndErrors.put("errorCode", errorCode);
			ratesAndErrors.put("errorMessage", responseStatus
					.get("errorDescription"));
			if (log.isDebugEnabled()) {
				log.debug("Error contacting UPS, information follows:");
				log.debug((String) responseStatus.get("errorCode"));
				log.info((String) responseStatus.get("errorDescription"));
				log.debug((String) responseStatus.get("errorSeverity"));
				log.debug((String) responseStatus
						.get("errorLocationElementName"));
			}
		} else if (Integer
				.parseInt(responseStatus.get("statusCode").toString()) == 1) {
			// Good
			ratesAndErrors = getRatesFromXml(xmlString);
		} else {
			// Who knows
			errorCode = "Unknown";
			errorMessage = (String) getSettings().getValue("UPS_getRatesError");
			ratesAndErrors.put("rates", rates);
			ratesAndErrors.put("errorCode", errorCode);
			ratesAndErrors.put("errorMessage", errorMessage);
			if (log.isDebugEnabled()) {
				log.debug("Error contacting UPS, information follows:");
				log.debug("Status code not returned.");
			}
		}

		return ratesAndErrors;
	}

	public Map getRatesFromXml(String xml) throws Exception {
		Map ratesAndErrors = new HashMap();
		Collection rates = new ArrayList();

		Document doc = null;
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		InputSource is = new InputSource(new StringReader(xml));
		doc = builder.parse(is);

		Node n = doc.getDocumentElement().getFirstChild();
		while (n != null) {
			if (n.getNodeName().equals("RatedShipment")) {
				String serviceCode = "";
				String totalCharges = "";
				String currencyCode = "USD";
				String negotiatedTotalCharges = null;
				String negotiatedCurrencyCode = "USD";
				String deliveryDays = "";
				Node n2 = n.getFirstChild();
				while (n2 != null) {
					if (n2.getNodeName().equals("Service")) {
						Node n3 = n2.getFirstChild();
						while (n3 != null) {
							if (n3.getNodeName().equals("Code")) {
								serviceCode = n3.getFirstChild().getNodeValue();
							}
							n3 = n3.getNextSibling();
						}
					}
					if (n2.getNodeName().equals("TotalCharges")) {
						Node n3 = n2.getFirstChild();
						while (n3 != null) {
							if (n3.getNodeName().equals("MonetaryValue")) {
								totalCharges = n3.getFirstChild()
										.getNodeValue();
							}
							if (n3.getNodeName().equals("CurrencyCode")) {
								currencyCode = n3.getFirstChild()
										.getNodeValue();
							}
							n3 = n3.getNextSibling();
						}
					}
					if (n2.getNodeName().equals("NegotiatedRates")) {
						Node n3 = n2.getFirstChild();
						while (n3 != null) {
							if (n3.getNodeName().equals("NetSummaryCharges")) {
								Node n4 = n3.getFirstChild();
								while (n4 != null) {
									if (n4.getNodeName().equals("GrandTotal")) {
										Node n5 = n4.getFirstChild();
										while (n5 != null) {
											if (n5.getNodeName().equals("MonetaryValue")) {
												negotiatedTotalCharges = n5.getFirstChild().getNodeValue();
											}
											if (n5.getNodeName().equals("CurrencyCode")) {
												negotiatedCurrencyCode = n5.getFirstChild().getNodeValue();
											}
											n5 = n5.getNextSibling();
										}
									}
									n4 = n4.getNextSibling();
								}
							}
							n3 = n3.getNextSibling();
						}
					}
					if (n2.getNodeName().equals("GuaranteedDaysToDelivery")
							&& n2.getFirstChild() != null) {
						deliveryDays = n2.getFirstChild().getNodeValue();
					}
					n2 = n2.getNextSibling();
				}

				// Negotiated or retail?
				if (UPS_rateInformationNegotiatedRatesIndicator && negotiatedTotalCharges != null) {
					log.info("Negotiated rates indicator is on and a negotiated total was found: "
							+ negotiatedTotalCharges + " replacing retail total of " + totalCharges);
					totalCharges = negotiatedTotalCharges;
					if (negotiatedCurrencyCode != null) {
						currencyCode = negotiatedCurrencyCode;
					}
				}

				// We have a code and charge
				if (serviceCode.length() > 0 && totalCharges.length() > 0) {
					if (isInCommaSeparatedString(serviceCode,
							UPS_validServiceCodes)) {
						if (log.isDebugEnabled())
							log.debug("Adding rate: " + serviceCode);
						// Allowed method to show user
						HashMap o = new HashMap();
						o.put("code", serviceCode);
						o.put("name", UPS_MethodsMap.get(serviceCode));
						double totalChargesDouble = Double
								.parseDouble(totalCharges);
						o.put("priceDouble", new Double(totalChargesDouble));
						o.put("originalPriceDouble", new Double(
								totalChargesDouble));
						o.put("price", formatPrice(totalChargesDouble,
								currencyCode, getUser().getLocale()));
						o.put("originalPrice", formatPrice(totalChargesDouble,
								currencyCode, getUser().getLocale()));
						o.put("processorClass", "UPSShippingProcessor");
						o.put("processorName", "UPS");
						o.put("currencyCode", currencyCode);
						String estimatedDeliveryDate = doEstimatedDeliveryDate(deliveryDays);
						if (estimatedDeliveryDate != null) {
							o.put("DeliveryDate", estimatedDeliveryDate);
						}
						Collection discounts = new ArrayList();
						o.put("discounts", discounts);
						rates.add(o);
					} else {
						// Not allowed method to show user
					}
				}
			}
			n = n.getNextSibling();
		}

		ratesAndErrors.put("rates", rates);

		return ratesAndErrors;

	}

	public Map getResponseStatusFromXml(String xml) throws Exception {
		Map status = new HashMap();

		Document doc = null;
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		InputSource is = new InputSource(new StringReader(xml));
		doc = builder.parse(is);

		/*
		 * Get the response status
		 */
		int statusCode = 0;
		String errorCode = "";
		String errorDescription = "";
		String errorSeverity = "";
		String errorLocationElementName = "";
		boolean foundResponse = false;

		Node n = doc.getDocumentElement().getFirstChild();
		while (n != null) {
			// if (log.isInfoEnabled()) log.info(n.getNodeName());
			if (n.getNodeName().equals("Response")) {
				foundResponse = true;
				Node n2 = n.getFirstChild();
				while (n2 != null) {
					if (n2.getNodeName().equals("ResponseStatusCode")) {
						statusCode = Integer.parseInt(n2.getFirstChild()
								.getNodeValue());
					}
					if (n2.getNodeName().equals("Error")) {
						Node n3 = n2.getFirstChild();
						while (n3 != null) {
							if (n3.getNodeName().equals("ErrorSeverity")) {
								errorSeverity = n3.getFirstChild()
										.getNodeValue();
							}
							if (n3.getNodeName().equals("ErrorCode")) {
								errorCode = n3.getFirstChild().getNodeValue();
							}
							if (n3.getNodeName().equals("ErrorDescription")) {
								errorDescription = n3.getFirstChild()
										.getNodeValue();
							}
							if (n3.getNodeName().equals("ErrorLocation")) {
								Node n4 = n3.getFirstChild();
								while (n4 != null) {
									if (n4.getNodeName().equals(
											"ErrorLocationElementName")) {
										errorLocationElementName = n4
												.getFirstChild().getNodeValue();
									}
									n4 = n4.getNextSibling();
								}
							}
							n3 = n3.getNextSibling();
						}
					}
					n2 = n2.getNextSibling();
				}
			}
			if (foundResponse) {
				break;
			}
			n = n.getNextSibling();
		}

		if (foundResponse) {
			status.put("statusCode", new Integer(statusCode));
			status.put("errorCode", errorCode);
			status.put("errorDescription", errorDescription);
			status.put("errorSeverity", errorSeverity);
			status.put("errorLocationElementName", errorLocationElementName);
		} else {
			status.put("statusCode", "0");
			status.put("errorCode", "Return Error");
			status.put("errorDescription", (String) getSettings().getValue(
					"UPS_accessLicenseNumber"));
			status.put("errorSeverity", "Fatal");
			status.put("errorLocationElementName", "");
		}

		return status;
	}

	public boolean isInCommaSeparatedString(String value,
			String commaSeparatedString) {
		boolean result = false;
		String[] separatedValues = commaSeparatedString.split(",");
		for (int j = 0; j < separatedValues.length; j++) {
			String thisValue = separatedValues[j].trim();
			if (value.equals(thisValue)) {
				result = true;
				break;
			}
		}
		return result;
	}

	public String doEstimatedDeliveryDate(String deliveryDays) {

		// Don't do anything if there's not delivery day provided
		if (deliveryDays == null || deliveryDays.length() == 0) {
			return null;
		}

		// Figure the initial delivery date based on what UPS provided
		Calendar shipDate = Calendar.getInstance();
		shipDate.add(Calendar.DATE, Integer.parseInt(deliveryDays));

		// Add extra days if they have been defined in the settings
		String extraDays = (String) getSettings().getValue(
				"upsExtraDeliveryDays");
		if (extraDays != null && extraDays.length() > 0) {
			shipDate.add(Calendar.DATE, Integer.parseInt(extraDays));
		}

		// Push the delivery date to the next weekday - doesn't take into
		// account holidays
		if (shipDate.get(Calendar.DAY_OF_WEEK) == Calendar.SATURDAY) {
			shipDate.add(Calendar.DATE, 2);
		} else if (shipDate.get(Calendar.DAY_OF_WEEK) == Calendar.SUNDAY) {
			shipDate.add(Calendar.DATE, 1);
		}

		// Now format the date
		String monthString = Integer.toString(shipDate.get(Calendar.MONTH) + 1);
		if (monthString.length() == 1) {
			monthString = "0" + monthString;
		}
		String dayString = Integer
				.toString(shipDate.get(Calendar.DAY_OF_MONTH));
		if (dayString.length() == 1) {
			dayString = "0" + dayString;
		}
		String shipDateString = shipDate.get(Calendar.YEAR) + "-" + monthString
				+ "-" + dayString;
		if (log.isDebugEnabled())
			log.debug("shipDateString: " + shipDateString);

		return shipDateString;
	}

	public Collection doGlobalExtraCharge(Collection mergedRates) {
		UPS_extraCharge = (String) getSettings().getValue("UPS_extraCharge");
		if (mergedRates == null || mergedRates.isEmpty()
				|| UPS_extraCharge == null
				|| UPS_extraCharge.replaceAll("\\D", "").length() == 0) {
			return mergedRates;
		}
		String currencyCode = (String) getSettings().getValue("currencyCode");
		Iterator i = mergedRates.iterator();
		while (i.hasNext()) {
			Map mergedRate = (Map) i.next();
			Double mergedPriceDouble = (Double) mergedRate.get("priceDouble");
			if (mergedPriceDouble == null) {
				mergedPriceDouble = new Double(0);
			}
			double newPrice = mergedPriceDouble.doubleValue();
			try {
				String extraCharge = UPS_extraCharge;
				extraCharge = extraCharge.trim();
				if (extraCharge.endsWith("%")) {

					// Charge is a percentage
					extraCharge.replaceAll("\\D", "");
					double percentageCharge = Double.parseDouble(extraCharge);
					double doubleCharge = utils().roundForMoney(
							percentageCharge * newPrice).doubleValue();
					doubleCharge = doubleCharge / 100;
					newPrice = newPrice + doubleCharge;
				} else {

					// Charge is an amount
					extraCharge.replaceAll("\\D", "");
					newPrice = newPrice + Double.parseDouble(extraCharge);
				}
			} catch (NumberFormatException nfe) {
				if (log.isErrorEnabled())
					log.error("UPS extra charge format error: "
							+ UPS_extraCharge, nfe);
			}
			mergedRate.put("priceDouble", new Double(newPrice));
			mergedRate.put("originalPriceDouble", new Double(newPrice));
			mergedRate.put("price", formatPrice(newPrice, currencyCode,
					getUser().getLocale()));
			mergedRate.put("originalPrice", formatPrice(newPrice, currencyCode,
					getUser().getLocale()));
		}
		return mergedRates;
	}
}
