package com.softslate.commerce.businessobjects.tax;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.softslate.commerce.businessobjects.core.BaseBusinessProcessor;
import com.softslate.commerce.businessobjects.order.Order;
import com.softslate.commerce.businessobjects.order.OrderDiscount;
import com.softslate.commerce.businessobjects.order.OrderItem;
import com.softslate.commerce.daos.order.OrderGatewayDAO;

/**
 * Abstract class that provides methods useful for <code>TaxProcessor</code>
 * implementations.
 * 
 * @author David Tobey
 * 
 */
public abstract class BaseTaxProcessor extends BaseBusinessProcessor implements
		TaxProcessor {

	static Log log = LogFactory.getLog(BaseTaxProcessor.class);

	public abstract int processTax() throws Exception;

	/**
	 * Assists with processing of tax charges. Copies the incoming
	 * <code>Map</code> of parameters to this <code>BaseTaxProcessor</code>
	 * instance and calls <code>updateTaxableSubtotal()</code>, followed by
	 * <code>processTax()</code>, followed by
	 * <code>updateTotalsAndStatuses()</code> to update the order's totals.
	 * Finally, writes the new totals to the database using
	 * {@link com.softslate.commerce.daos.order.OrderGatewayDAO#updateOrderAndDelivery}.
	 */
	public Map processTax(Map parameters) throws Exception {
		if (log.isInfoEnabled())
			log.info("Processing tax.");
		Map results = new HashMap();
		int result = 0;
		try {
			// Copy the incoming parameters to this object.
			BeanUtils.copyProperties(this, parameters);
			if (log.isDebugEnabled())
				log.debug("Updating taxable subtotal.");
			updateTaxableSubtotal();
			if (log.isDebugEnabled())
				log.debug("Setting tax costs.");
			result = processTax();
			if (log.isDebugEnabled())
				log.debug("Updating totals and statuses.");
			updateOrderTotals();
			if (getUser().getOrder().getCompleted() == null) {
				getUser().getOrder().setStatus("Tax processed");
			}
			OrderGatewayDAO orderGatewayDAO =  getDaoFactory()
					.createDAO(OrderGatewayDAO.class);
			orderGatewayDAO.updateOrderAndDelivery(true, getUser().getOrder(),
					getUser().getOrder().firstDelivery());

		} catch (Exception e) {
			if (getUser().getOrder().getCompleted() == null) {
				getUser().getOrder().setStatus("Corrupted");
			}
			throw e;
		}
		results.put("resultCode", Integer.toString(result));
		return results;
	}

	/**
	 * Computes and updates the taxable subtotal of a user's order. The taxable
	 * subtotal is determined by first looping through each of the cart items.
	 * The total of each item whose corresponding product is taxable is added to
	 * the subtotal. Next, the current order's discounts are looped over. The
	 * totals of discounts corresponding to order items that are taxable are
	 * subtracted from the taxable subtotal. Global discounts are also
	 * subtracted, if the "applyDiscountsToTaxableSubtotal" setting is set to
	 * true.
	 * <p>
	 * The user's <code>Order</code> and that order's first
	 * <code>OrderDelivery</code> are updated with the new taxable subtotal.
	 * 
	 * @throws Exception
	 */
	public void updateTaxableSubtotal() throws Exception {
		updateTaxableSubtotal(getUser().getOrder());
	}
	public void updateTaxableSubtotal(Order order) throws Exception {
		double taxableSubtotal = 0;
		// Loop through the order items
		Collection ois = order.firstDelivery()
				.getOrderItemCollection();
		Iterator i = ois.iterator();
		while (i.hasNext()) {
			OrderItem oi = (OrderItem) i.next();
			if (oi.getIsTaxed()) {
				taxableSubtotal = taxableSubtotal + oi.getTotal().doubleValue();
			}
		}
		// Loop through the order discounts
		if (order.getOrderDiscounts() != null) {
			String applyDiscountsToTaxableSubtotal = (String) getSettings()
					.getValue("applyDiscountsToTaxableSubtotal");
			Collection ods = order.getOrderDiscounts();
			Iterator j = ods.iterator();
			while (j.hasNext()) {
				OrderDiscount od = (OrderDiscount) j.next();
				if (od.getItemsAppliesType().equals("[skus]")
						&& od.getOrderItem() != null
						&& od.getOrderItem().getIsTaxed()) {
					taxableSubtotal = taxableSubtotal
							- od.getTotal().doubleValue();
				} else if (od.getItemsAppliesType().equals("[global]")
						&& applyDiscountsToTaxableSubtotal != null
						&& applyDiscountsToTaxableSubtotal.equals("1")) {
					taxableSubtotal = taxableSubtotal
							- od.getTotal().doubleValue();
				}
			}
		}
		// Update taxable subtotal depending on if shipping is taxable.
		boolean isStateWhereShippingIsTaxable = isStateWhereShippingIsTaxable(order);
		if (isStateWhereShippingIsTaxable) {
			Double dShip = order.firstDelivery().getShipping();
			if (dShip != null) {
				taxableSubtotal = taxableSubtotal + dShip.doubleValue();
			}
		}
		// Make sure the taxable subtotal is not less than 0
		if (taxableSubtotal < 0) {
			taxableSubtotal = 0;
		}
		// Set the taxable subtotals of the order and the first delivery
		order.setTaxableSubtotal(new Double(taxableSubtotal));
		order.firstDelivery().setTaxableSubtotal(
				new Double(taxableSubtotal));
	}

	protected boolean isStateWhereShippingIsTaxable(Order order) {
		String statesWhereShippingIsTaxable = (String) getSettings().getValue("statesWhereShippingIsTaxable");
		boolean isUSOrCanada = order != null && order.firstDelivery() != null
				&& order.firstDelivery().getCountry() != null
				&& (order.firstDelivery().getCountry().equals("US") || order.firstDelivery().getCountry().equals("CA"));
		boolean isStateWhereShippingIsTaxable = isUSOrCanada && statesWhereShippingIsTaxable != null
				&& statesWhereShippingIsTaxable.length() > 0 && order != null && order.firstDelivery() != null
				&& order.firstDelivery().getState() != null && order.firstDelivery().getState().length() > 0
				&& statesWhereShippingIsTaxable.contains(order.firstDelivery().getState());
		if (isStateWhereShippingIsTaxable) {
			log.info("User's state is set to tax shipping.");
			if (order != null && order.firstDelivery() != null) {
				log.info("User's state: " + order.firstDelivery().getState());
			}
		}
		return isStateWhereShippingIsTaxable;
	}

}
