package com.softslate.commerce.customer.core;

import java.util.Map;

import javax.servlet.http.HttpServletRequest;

import com.softslate.commerce.businessobjects.core.Settings;
import com.softslate.commerce.businessobjects.product.SearchList;

/**
 * Interface for utility methods used by Struts Form classes.
 * <p>
 * When the application needs to create an instance that implements
 * <code>FormUtils</code>, it calls
 * {@link com.softslate.commerce.businessobjects.core.BusinessObjectFactory#createUtilsObject} ,
 * which finds the name of the Java class to instantiate from the
 * "formUtilsImplemeter" setting in the <code>appComponents.properties</code>
 * file.
 * <p>
 * The default "formUtilsImplemeter" is
 * {@link com.softslate.commerce.customer.core.FormUtilsImpl}.
 */
public interface FormUtils {

	/**
	 * Determines if a given parameter was included in the request for a
	 * <code>BaseDynaForm</code> and is of non-zero length.
	 * 
	 * @param form
	 *            The <code>BaseDynaForm</code> being processed.
	 * @param property
	 *            The name of the parameter being tested.
	 * @return <code>true</code> if the given parameter was submitted to the
	 *         form and is of non-zero length. <code>false</code> if not.
	 */
	abstract boolean stringExists(BaseDynaForm form, String property);

	/**
	 * Determines if two string parameters submitted to a
	 * <code>BaseDynaForm</code> exist and are equal.
	 * 
	 * @param form
	 *            The <code>BaseDynaForm</code> being processed.
	 * @param property1
	 *            The first parameter to be tested.
	 * @param property2
	 *            The second parameter to be tested.
	 * @return <code>true</code> if the given parameters were submitted to the
	 *         form and are equal. <code>false</code> if not.
	 */
	abstract boolean twoPropertiesEqual(BaseDynaForm form, String property1,
			String property2);

	/**
	 * Determines if a given parameter for a <code>BaseDynaForm</code> is
	 * greater than a given <code>int</code>.
	 * 
	 * @param form
	 *            The <code>BaseDynaForm</code> being processed.
	 * @param property
	 *            The name of the parameter to be tested.
	 * @param value
	 *            The value for the parameter to be tested against.
	 * @return <code>true</code> if the given parameter is greater than the
	 *         value. <code>false</code> if not.
	 */
	abstract boolean stringValueGreaterThan(BaseDynaForm form, String property,
			int value);

	/**
	 * Determines if a given parameter for a <code>BaseDynaForm</code> is less
	 * than a given <code>int</code>.
	 * 
	 * @param form
	 *            The <code>BaseDynaForm</code> being processed.
	 * @param property
	 *            The name of the parameter to be tested.
	 * @param value
	 *            The value for the parameter to be tested against.
	 * @return <code>true</code> if the given parameter is less than the
	 *         value. <code>false</code> if not.
	 */
	abstract boolean stringValueLessThan(BaseDynaForm form, String property,
			int value);

	/**
	 * 
	 * Determines if at least one element was submitted and is of non-zero
	 * length for a string array parameter of a <code>BaseDynaForm</code>.
	 * 
	 * @param form
	 *            The <code>BaseDynaForm</code> being processed.
	 * @param property
	 *            The name of the parameter being tested.
	 * @return <code>true</code> if the given parameter has at least one
	 *         element and is of non-zero length. <code>false</code> if not.
	 */
	abstract boolean stringArrayHasAnElement(BaseDynaForm form, String property);

	/**
	 * Determines if at least one element was submitted and is greater than a
	 * given <code>int</code> for a string array parameter of a
	 * <code>BaseDynaForm</code>.
	 * 
	 * @param form
	 *            The <code>BaseDynaForm</code> being processed.
	 * @param property
	 *            The name of the parameter being tested.
	 * @param value
	 *            The value for the parameter to be tested against.
	 * @return <code>true</code> if the given parameter has at least one
	 *         element and is greater than the value. <code>false</code> if
	 *         not.
	 */
	abstract boolean stringArrayValuesGreaterThan(BaseDynaForm form,
			String property, int value);

	/**
	 * Determines if a given string array has at least one element and that it's
	 * of non-zero length.
	 * 
	 * @param paramValues
	 *            The string array being tested.
	 * @return <code>true</code> if the given string array has one element
	 *         that's of non-zero length.
	 */
	abstract boolean stringArrayHasAnElement(String[] paramValues);

	/**
	 * Determines if at least one element in a string array is greater than a
	 * given value.
	 * 
	 * @param paramValues
	 *            The string array being tested.
	 * @param value
	 *            The value for the string array's elements to be tested
	 *            against.
	 * @return <code>true</code> if at least one element in a string array is
	 *         greater than a given value. <code>false</code> if not.
	 */
	abstract boolean stringArrayValuesGreaterThan(String[] paramValues,
			int value);

	/**
	 * Returns a new string with the first letter of a given string capitalized.
	 * 
	 * @param stringToCapitalize
	 *            The string to capitalize.
	 * @return A new string with the first letter capitalized.
	 */
	abstract String capitalize(String stringToCapitalize);

	/**
	 * Checks to see if a given string is in the form of a valid email address.
	 * Uses the following regular expression for the test:
	 * <code>.+@.+\\.[a-z]+</code>
	 * 
	 * @param email
	 *            The string to be tested.
	 * @return <code>true</code> if the given string matches a valid email
	 *         address.
	 */
	abstract boolean isValidEmail(String email);

	/**
	 * Checks to see if a given string is in the form of a valid phone number.
	 * Since phone numbers may be U.S. or international in format, and may or
	 * may not have an extension, only a light check is done ensuring that at
	 * least 10 digits are present.
	 * 
	 * @param phone
	 *            The string to be tested.
	 * @return <code>true</code> if the given string matches a valid phone
	 *         number.
	 */
	abstract boolean isValidPhone(String phone);

	/**
	 * Checks to see if a given string is in the form of a valid postal code,
	 * given an associated country. If the country is US, a simple check on the
	 * number of digits is done.
	 * 
	 * @param postalCode
	 *            The string to be tested.
	 * @return <code>true</code> if the given string matches a valid postal
	 *         code.
	 */
	abstract boolean isValidPostalCode(String postalCode, String country);

	/**
	 * Checks to see if a given string is in the form of a valid integer.
	 * 
	 * @param value
	 *            The string to be tested.
	 * @return <code>true</code> if the given string is a valid integer
	 */
	abstract boolean isInteger(String value);

	/**
	 * Runs a given string through the Luhn algorithm to determine if it is a
	 * valid credit card number.
	 * 
	 * @param creditCardNumber
	 *            The credit card number
	 * @return true, if the string is a valid credit card number; false,
	 *         otherwise.
	 */
	abstract boolean isCreditCardValid(String creditCardNumber);

	/**
	 * Given a credit card type and a number, checks that the number corresponds
	 * correctly with the type.
	 * 
	 * @param type
	 *            Four types supported: "VISA", "DISCOVER", "AMEX", and
	 *            "MASTERCARD"
	 * @param number
	 *            The credit card number
	 * @return true if the card type matches the number or if the card type is
	 *         not supported.
	 */
	abstract boolean doCardTypeAndNumberMatch(String type, String number);

	/**
	 * Given a credit card number, derive its type (Visa, Discover, MasterCard
	 * or AMEX).
	 * 
	 * @param number
	 *            The credit card number
	 * @return Four types supported: "VISA", "DISCOVER", "AMEX", and
	 *         "MASTERCARD"
	 */
	abstract String deriveCardType(String number);

	/**
	 * Determines if a given string contains a character residing in another
	 * string.
	 * 
	 * @param string
	 *            The string to check against
	 * @param characters
	 *            A String containing each of the characters to check
	 * @return the first matching Character, if the string contains one of the
	 *         characters, otherwise null
	 */
	abstract Character containsCharacter(String string, String characters);

	abstract String stringArrayAsURLParam(String parameterName, String[] stringArray) throws Exception;

	abstract void eliminateDuplicateFacets(SearchList searchList);

	abstract void removePrefixedKeysFromMap(Map properties, String string);

	abstract void removeNullsFromMap(Map incomingProperties);

	abstract void removeAllNotInRequestFromMap(Map incomingProperties, HttpServletRequest request);

	abstract String html5OrXhtml(HttpServletRequest request, Settings settings);

}
