package com.softslate.commerce.customer.order;

import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Currency;
import java.util.Iterator;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;

import com.softslate.commerce.businessobjects.order.CartDiscountProcessor;
import com.softslate.commerce.businessobjects.order.OrderDiscount;
import com.softslate.commerce.customer.core.BaseForm;

/**
 * Struts action class for submissions to record a coupon code entered by the
 * user. Used for requests to "/CartCoupon.do" and "/CheckoutCoupon.do".
 * <p>
 * The <code>execute</code> method of this class first adds the submitted
 * coupon code string to the <code>enteredCouponCodes</code> property of the
 * current user's {@link com.softslate.commerce.businessobjects.core.User}
 * instance. Next, it reprocesses discounts by invoking
 * {@link com.softslate.commerce.businessobjects.order.CartDiscountProcessor#processCouponCode(Map)}.
 * Next, it checks the user's applied discounts to see if the coupon was
 * applied. If it was not applied, an error is added to the request, the coupon
 * code is <em>removed</em> from the <code>enteredCouponCodes</code>
 * property, and the <code>ActionForward</code> returned will be named
 * "failureCouponNotApplied". If the coupon is successfully applied, the
 * "success" <code>ActionForward</code> is returned.
 * 
 * @author David Tobey
 */
public class CouponAction extends CheckoutAction {
	static Log log = LogFactory.getLog(CouponAction.class);

	public ActionForward execute(ActionMapping mapping, ActionForm form,
			HttpServletRequest request, HttpServletResponse response)
			throws Exception {
		if (log.isInfoEnabled())
			log.info("Starting action.");

		// Initialize the BaseForm
		BaseForm baseForm = new BaseForm();
		baseForm.initializeProperties(mapping, request);
		baseForm.setServlet(getServlet());

		synchronized (baseForm.getRequest().getSession()) {
			// Process the coupon.
			String couponCode = baseForm.getRequest()
					.getParameter("couponCode");
			if (couponCode == null || couponCode.trim().length() == 0
					|| baseForm.getUser().getOrder() == null) {
				if (request.getAttribute("isApiRequest") != null && (Boolean) request.getAttribute("isApiRequest")) {
					baseForm.getMessages().add(ActionMessages.GLOBAL_MESSAGE,
							new ActionMessage("messages.noCouponCodeOrNoOrder"));
					saveMessages(baseForm.getRequest(), baseForm.getMessages());
				}
				return baseForm.getMapping().findForward("success");
			}
			String applyDiscounts = (String) baseForm.getSettings().getValue(
					"applyDiscounts");
			if (applyDiscounts == null || applyDiscounts.equals("0")) {
				return baseForm.getMapping().findForward("success");
			}
			couponCode = couponCode.trim();
			String escapedCouponCode = StringEscapeUtils.escapeXml11(couponCode);
			if (log.isDebugEnabled())
				log.debug("Processing coupon code " + couponCode);
			
			// Add the coupon code to the collection of entered coupon codes.
			String allowStackedCoupons = (String) baseForm.getSettings()
			.getValue("allowStackedCoupons");
			if (baseForm.getUser().getEnteredCouponCodes() == null
					|| (allowStackedCoupons == null || allowStackedCoupons
							.equals("0"))) {
				baseForm.getUser().setEnteredCouponCodes(new ArrayList());
			}
			baseForm.getUser().getEnteredCouponCodes().add(couponCode);
			// Process discounts
			CartDiscountProcessor dp = (CartDiscountProcessor) baseForm
					.getBusinessObjectFactory().createObject(
							"cartDiscountProcessorImplementer");
			Map parameters = PropertyUtils.describe(baseForm);
			parameters.put("couponCode", couponCode);
			dp.processCouponCode(parameters);
			// Force the shipping and taxes to be reprocessed
			baseForm.getRequest().getSession()
					.setAttribute("shippingAndTaxesProcessed", "false");
			
			// Loop through the user's discounts to see if a matching discount
			// was added
			Collection ods = baseForm.getUser().getOrder().getOrderDiscounts();
			if (ods != null && !ods.isEmpty()) {
				Iterator i = ods.iterator();
				while (i.hasNext()) {
					OrderDiscount od = (OrderDiscount) i.next();
					if (od.getCouponCode() != null
							&& od.getCouponCode().equalsIgnoreCase(couponCode)) {
						if (log.isInfoEnabled())
							log.info("Coupon code " + od.getCouponCode()
									+ " applied");
						if (od.getItemsAppliesType() != null
								&& od.getItemsAppliesType().equals(
										"[giftCertificate]")) {
							String fb = od.getReward().toString();
							try {
								NumberFormat nf = NumberFormat
										.getCurrencyInstance(baseForm.getUser()
												.getLocale());
								String cc = (String) baseForm.getSettings()
										.getValue("currencyCode");
								nf.setCurrency(Currency.getInstance(cc));
								fb = nf.format(od.getReward().doubleValue());
							} catch (Exception e) {
								if (log.isErrorEnabled()) {
									log.error("Error formatting balance: ", e);
								}
							}
							baseForm.getMessages().add(
									ActionMessages.GLOBAL_MESSAGE,
									new ActionMessage(
											"messages.giftCertWillBeApplied",
											escapedCouponCode, fb));
						} else {
							baseForm.getMessages().add(
									ActionMessages.GLOBAL_MESSAGE,
									new ActionMessage("messages.couponApplied",
											escapedCouponCode));
						}
						saveMessages(baseForm.getRequest(),
								(ActionMessages) baseForm.getMessages());
						return baseForm.getMapping().findForward(
								"successCouponApplied");
					}
				}
			}

			// Coupon code not applied - tell the user and send back an error
			if (log.isDebugEnabled())
				log.debug("Coupon code " + couponCode + " not applicable");
			baseForm.getUser().getEnteredCouponCodes().remove(couponCode);
			baseForm.getErrors().add(ActionMessages.GLOBAL_MESSAGE,
					new ActionMessage("errors.couponNotApplied", escapedCouponCode));

			// Check if there is a "couponNotAppliedReasons" matching the coupon that was entered. In this case, we'll
			// tell the customer why it was not applied.
			if (baseForm.getUser().getCustomMap() != null
					&& baseForm.getUser().getCustomMap().get("couponNotAppliedReasons") != null) {
				String reasonCode = ((Map<String, String>) baseForm.getUser().getCustomMap()
						.get("couponNotAppliedReasons")).get(couponCode);
				String errorKey = "errors." + reasonCode;
				if (StringUtils.isNotBlank(reasonCode)) {
					String message = baseForm.getMessage("core", errorKey, null);
					if (StringUtils.isNotBlank(message)) {
						baseForm.getErrors().add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage(errorKey));
					} else {
						baseForm.getErrors().add(ActionMessages.GLOBAL_MESSAGE,
								new ActionMessage("errors.detail", errorKey));
					}
				}
			}
			saveErrors(baseForm.getRequest(),
					(ActionMessages) baseForm.getErrors());
			return baseForm.getMapping().findForward("failureCouponNotApplied");
		}
	}
}
