package com.softslate.commerce.daos.core;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.SessionFactory;

import com.softslate.commerce.businessobjects.core.Settings;

/**
 * @author David Tobey
 */
public abstract class BaseDAOFactory implements DAOFactory {

	static Log log = LogFactory.getLog(BaseDAOFactory.class);

	/**
	 * A <code>Map</code> of SQL statements used by the application.
	 */
	private Map sqlMap = null;

	public Map getSqlMap() {
		return (this.sqlMap);
	}

	public void setSqlMap(Map sqlMap) {
		this.sqlMap = sqlMap;
	}

	/**
	 * The list of modules in use by the application in
	 * <code>StringTokenizer</code> form. Modules may include core, product,
	 * order, and customer.
	 */
	private StringTokenizer moduleList = null;

	public StringTokenizer getModuleList() {
		return (this.moduleList);
	}

	public void setModuleList(StringTokenizer moduleList) {
		this.moduleList = moduleList;
	}

	/**
	 * The list of modules in use by the application in <code>String</code>
	 * form. Modules may include core, product, order, and customer.
	 */
	public static String moduleListString = "core,product,order,customer,administrator,tax,shipping,importexport,payment,quartz,email,envers";

	public String getModuleListString() {
		return (moduleListString);
	}

	public void setModuleListString(String moduleList) {
		moduleListString = moduleList;
	}

	/**
	 * The application's <code>DataSource</code> used for persistent storage.
	 */
	private DataSource dataSource = null;

	public DataSource getDataSource() {
		return (this.dataSource);
	}

	public void setDataSource(DataSource dataSource) {
		this.dataSource = dataSource;
	}

	/**
	 * The JNDI name for the application's <code>DataSource</code>, if one is
	 * set up in the servlet container's configuration. (For example, in
	 * server.xml for Tomcat.)
	 */
	private String dataSourceName = null;

	public String getDataSourceName() {
		return (this.dataSourceName);
	}

	public void setDataSourceName(String dataSourceName) {
		this.dataSourceName = dataSourceName;
	}

	/**
	 * The JNDI context for the application's <code>DataSource</code>, if one
	 * is set up in the servlet container's configuration. (For example, in
	 * server.xml for Tomcat.)
	 */
	private String jndiContext = null;

	public String getJndiContext() {
		return (this.jndiContext);
	}

	public void setJndiContext(String jndiContext) {
		this.jndiContext = jndiContext;
	}

	/**
	 * The fully-qualified class name of the database driver used for the
	 * <code>DataSource</code>.
	 */
	private String databaseDriver = null;

	public String getDatabaseDriver() {
		return (this.databaseDriver);
	}

	public void setDatabaseDriver(String databaseDriver) {
		this.databaseDriver = databaseDriver;
	}

	/**
	 * The URL used for this factory's <code>DataSource</code> database
	 * connection, if the built-in connection pooling is used. This factory will
	 * use its own connection pooling if it fails to connect using the JNDI
	 * <code>dataSourceName</code> and <code>jndiContext</code> settings.
	 */
	private String databaseURL = null;

	public String getDatabaseURL() {
		return (this.databaseURL);
	}

	public void setDatabaseURL(String databaseURL) {
		this.databaseURL = databaseURL;
	}

	/**
	 * The user name used for this factory's <code>DataSource</code> database
	 * connection, if the built-in connection pooling is used. This factory will
	 * use its own connection pooling if it fails to connect using the JNDI
	 * <code>dataSourceName</code> and <code>jndiContext</code> settings.
	 */
	private String databaseUserName = null;

	public String getDatabaseUserName() {
		return (this.databaseUserName);
	}

	public void setDatabaseUserName(String databaseUserName) {
		this.databaseUserName = databaseUserName;
	}

	/**
	 * The password used for this factory's <code>DataSource</code> database
	 * connection, if the built-in connection pooling is used. This factory will
	 * use its own connection pooling if it fails to connect using the JNDI
	 * <code>dataSourceName</code> and <code>jndiContext</code> settings.
	 */
	private String databasePassword = null;

	public String getDatabasePassword() {
		return (this.databasePassword);
	}

	public void setDatabasePassword(String databasePassword) {
		this.databasePassword = databasePassword;
	}

	/**
	 * The type of database this factory connects to. For example, "MSSQL" or
	 * "MySQL".
	 */
	private String databaseType = null;

	public String getDatabaseType() {
		return (this.databaseType);
	}

	public void setDatabaseType(String databaseType) {
		this.databaseType = databaseType;
	}

	/**
	 * The maximum number of idle connections for the pool used by this
	 * factory's <code>DataSource</code>, if the built-in connection pooling
	 * is used. This factory will use its own connection pooling if it fails to
	 * connect using the JNDI <code>dataSourceName</code> and
	 * <code>jndiContext</code> settings.
	 */
	private String maxIdle = null;

	public String getMaxIdle() {
		return this.maxIdle;
	}

	public void setMaxIdle(String maxIdle) {
		this.maxIdle = maxIdle;
	}

	/**
	 * The maximum number of active connections for the pool used by this
	 * factory's <code>DataSource</code>, if the built-in connection pooling
	 * is used. This factory will use its own connection pooling if it fails to
	 * connect using the JNDI <code>dataSourceName</code> and
	 * <code>jndiContext</code> settings.
	 */
	private String maxActive = null;

	public String getMaxActive() {
		return this.maxActive;
	}

	public void setMaxActive(String maxActive) {
		this.maxActive = maxActive;
	}

	/**
	 * The maximum time to wait for a connection to become available in
	 * milliseconds, before an exception is thrown. Used by this factory's
	 * <code>DataSource</code>, if the built-in connection pooling is used.
	 * This factory will use its own connection pooling if it fails to connect
	 * using the JNDI <code>dataSourceName</code> and <code>jndiContext</code>
	 * settings.
	 */
	private String maxWait = null;

	public String getMaxWait() {
		return this.maxWait;
	}

	public void setMaxWait(String maxWait) {
		this.maxWait = maxWait;
	}

	/**
	 * If set to true, abandoned connections will be removed and recycled. Used
	 * by this factory's <code>DataSource</code>, if the built-in connection
	 * pooling is used. This factory will use its own connection pooling if it
	 * fails to connect using the JNDI <code>dataSourceName</code> and
	 * <code>jndiContext</code> settings.
	 */
	private String removeAbandoned = null;

	public String getRemoveAbandoned() {
		return this.removeAbandoned;
	}

	public void setRemoveAbandoned(String removeAbandoned) {
		this.removeAbandoned = removeAbandoned;
	}

	/**
	 * The number of seconds a connection is idle before it is considered
	 * abandoned. Used by this factory's <code>DataSource</code>, if the
	 * built-in connection pooling is used. This factory will use its own
	 * connection pooling if it fails to connect using the JNDI
	 * <code>dataSourceName</code> and <code>jndiContext</code> settings.
	 */
	private String removeAbandonedTimeout = null;

	public String getRemoveAbandonedTimeout() {
		return this.removeAbandonedTimeout;
	}

	public void setRemoveAbandonedTimeout(String removeAbandonedTimeout) {
		this.removeAbandonedTimeout = removeAbandonedTimeout;
	}

	/**
	 * The full path to the file on the server where the private key is stored
	 * for key pair encryption.
	 */
	private String paymentPrivateKeyFile = null;

	public String getPaymentPrivateKeyFile() {
		return (this.paymentPrivateKeyFile);
	}

	public void setPaymentPrivateKeyFile(String paymentPrivateKeyFile) {
		this.paymentPrivateKeyFile = paymentPrivateKeyFile;
	}

	/**
	 * The full path to the file on the server where the public key is stored
	 * for key pair (RSA) encryption.
	 */
	private String paymentPublicKeyFile = null;

	public String getPaymentPublicKeyFile() {
		return (this.paymentPublicKeyFile);
	}

	public void setPaymentPublicKeyFile(String paymentPublicKeyFile) {
		this.paymentPublicKeyFile = paymentPublicKeyFile;
	}

	/**
	 * The full path to the file on the server where the key is stored for two
	 * way (DES) encryption.
	 */
	private String twoWayKeyFile = null;

	public String getTwoWayKeyFile() {
		return (this.twoWayKeyFile);
	}

	public void setTwoWayKeyFile(String twoWayKeyFile) {
		this.twoWayKeyFile = twoWayKeyFile;
	}

	/**
	 * Used to look up the Java class names for the various DAO's this factory
	 * creates.
	 */
	private Settings settings = null;

	public Settings getSettings() {
		return (this.settings);
	}

	public void setSettings(Settings settings) {
		this.settings = settings;
	}

	/**
	 * Initializes this <code>DAOFactory</code>. Called by
	 * <code>BaseRequestProcessor</code>.
	 * 
	 * @throws Exception
	 */
	public void initialize() throws Exception {
		return;
	}

	/**
	 * Initializes this <code>DAOFactory</code> using the settings defined in
	 * the file located at the imcoming <code>settingLocation</code>.
	 * 
	 * @param settingsLocation
	 *            The location of a file containing database settings.
	 * @throws Exception
	 */
	public void initialize(String settingsLocation) throws Exception {
		return;
	}

	/**
	 * Reads in the application's SQL statements into this factory's
	 * <code>sqlMap</code>. Loops through each of the active modules defined
	 * in the <code>moduleList</code> property and reads in the SQL
	 * properities from the files in "/WEB-INF/classes/resources.
	 * 
	 * @throws Exception
	 */
	public void initializeSQLProperties() throws Exception {
		return;
	}

	/**
	 * Loads database settings from a properties file into the properties of
	 * this <code>DAOFactory</code>.
	 * 
	 * @param settingsLocation
	 *            The location of a file containing database settings.
	 */
	public void initializeSettings(String settingsLocation) {
		return;
	}

	/**
	 * Loads SQL statements into a <code>Properties</code> object for a given
	 * database type. Reads a default sql.properties file for a given module
	 * under "WEB-INF/classes/resources" and then another file with SQL
	 * statements that override the defaults with SQL specific to the given
	 * database type. Finally, loads the sql-custom.properties file in each
	 * module to provide a way to override any of the statements deployed with
	 * the application.
	 * 
	 * @param databaseType
	 *            The type of database the SQL statements apply to.
	 * @param module
	 *            The module the SQL statements correspond to. For example,
	 *            core, product, order, or customer.
	 * @return A <code>Properties</code> object representing the SQL
	 *         statements.
	 * @throws Exception
	 */
	public Properties getProperties(String databaseType, String module)
			throws Exception {
		Properties returnedProperties = null;
		Properties databaseProperties = null;
		Properties customProperties = null;
		String pathToDefault = "resources/" + module + "/sql.properties";
		String pathToDBProperties = "resources/" + module + "/sql_"
				+ getDatabaseType().toLowerCase() + ".properties";
		String pathToCustom = "resources/" + module + "/sql-custom.properties";
		returnedProperties = loadProperties(pathToDefault);
		databaseProperties = loadProperties(pathToDBProperties,
				returnedProperties);
		if (databaseProperties != null) {
			if (log.isInfoEnabled())
				log.info("Loading sql statements for module: " + module
						+ "; database type: " + databaseType);
			returnedProperties = databaseProperties;
		}
		customProperties = loadProperties(pathToCustom, returnedProperties);
		if (customProperties != null) {
			if (log.isInfoEnabled())
				log.info("Loading sql-custom.properties file for module: "
						+ module);
			returnedProperties = customProperties;
		}
		return returnedProperties;
	}

	/**
	 * 
	 * Load properties from a .properties file represented by the argument.
	 * 
	 * @param path
	 *            The path to the .properties file being loaded.
	 * @return A <code>Properties</code> object representing the contents of
	 *         the file.
	 */
	public Properties loadProperties(String path) {
		InputStream is = null;
		is = getClass().getClassLoader().getResourceAsStream(path);
		Properties properties = null;
		BufferedInputStream bis = new BufferedInputStream(is);
		try {
			properties = new Properties();
			properties.load(bis);
			bis.close();
			is.close();
		} catch (IOException e) {
			properties = null;
			log.error(e.toString(),e);
		} finally {
			is = null;
			bis = null;
		}
		return properties;
	}

	/**
	 * Load properties from a .properties file represented by the argument,
	 * using a given <code>Properties</code> object as the default properties.
	 * 
	 * @param path
	 *            The path to the .properties file being loaded.
	 * @param defaultProperties
	 *            A <code>Properties</code> object representing the default
	 *            properties to be returned.
	 * @return A <code>Properties</code> object representing the contents of
	 *         the file.
	 */
	public Properties loadProperties(String path, Properties defaultProperties) {
		InputStream is = null;
		is = getClass().getClassLoader().getResourceAsStream(path);
		Properties properties = null;
		BufferedInputStream bis = new BufferedInputStream(is);
		try {
			properties = new Properties(defaultProperties);
			properties.load(bis);
			bis.close();
			is.close();
		} catch (IOException e) {
			properties = null;
			log.error(e.toString(), e);
		} finally {
			is = null;
			bis = null;
		}
		return properties;
	}

	/**
	 * Initializes this factory's <code>dataSource</code> property. This
	 * method first attempts to establish a <code>DataSource</code> based on a
	 * JNDI data source name and context defined in the JNDI environment, using
	 * the <code>datasourceName</code> and <code>jndiContext</code>
	 * properties. If this fails, it will create it's own connection pool using
	 * the various other properties defined in this factory, including the
	 * <code>databaseURL</code>,<code>databaseUserName</code>, and
	 * <code>databasePassword</code> properties.
	 */
	public void initializeDataSource() throws Exception {
		return;
	}

	/**
	 * Looks up a <code>DataSource</code> object in a JNDI context. Attempts
	 * to establish a <code>DataSource</code> based on a JNDI data source name
	 * and context defined in the JNDI environment, using the
	 * <code>datasourceName</code> and <code>jndiContext</code> properties.
	 * 
	 * @return A <code>DataSource</code> object for the application.
	 */
	public DataSource getJndiDataSource() throws Exception {
		return null;
	}

	/**
	 * Creates a connection pool using various properties defined in this
	 * factory. Returns a <code>DataSource</code> object tied to it.
	 * 
	 * @return A <code>DataSource</code> object for the application.
	 */
	public DataSource getPooledDataSource() throws Exception {
		return null;
	}

	/**
	 * Retrieves a <code>Map</code> of properties that will be transfered to
	 * each of the data access objects this factory creates.
	 * 
	 * @return A <code>Map</code> of properties that will be transfered to
	 *         each of the data access objects this factory creates
	 */
	public Map getTransferMap() {
		return null;
	}

	/**
	 * Creates and initializes a settings bean data access object for use by the
	 * application.
	 * 
	 * @return An instantiated and initialized settings bean data access object,
	 *         or <code>null</code> if initialization failed.
	 * @throws Exception
	 */
	public Object createSettingsDAO() throws Exception {
		return null;
	}

	/**
	 * Creates and initializes a data access object for use by the application.
	 * 
	 * @param type
	 *            A <code>String</code> associated with the class name of the
	 *            data access object to be created.
	 * @return An instantiated and initialized data access object, or
	 *         <code>null</code> if initialization failed.
	 * @throws Exception
	 */
	public Object createDAO(String type) throws Exception {
		return null;
	}

	/**
	 * Creates and initializes a data access object for use by the application.
	 * Using the incoming class name, appends the value of the
	 * <code>databaseType</code> property to the class name and attempts to
	 * instantiate that object. If that fails, it next appends "Basic" to the
	 * end of the class name. If a class still cannot be found, the class name
	 * from the class name itself is used.
	 * <p>
	 * For example, if this factory receives a request to create an instance of
	 * <code>CustomerDAO</code>, this factory will try to instantiate the
	 * following objects in this order:
	 * <ol>
	 * <li><code>com.softslate.commerce.daos.customer.CustomerDAOMSSQL</code>
	 * </li>
	 * <li><code>com.softslate.commerce.daos.customer.CustomerDAOBasic</code>
	 * </li>
	 * <li><code>com.softslate.commerce.daos.customer.CustomerDAO</code>
	 * </li>
	 * </ol>
	 * <p>
	 * Immediately after instantiating the requested DAO, this factory passes
	 * its own properties to it so that the newly created object can use them to
	 * communicate with the rest of the application.
	 * 
	 * @param className
	 *            A <code>String</code> representing the class name of the
	 *            data access object to be created.
	 * @return An instantiated and initialized data access object, or
	 *         <code>null</code> if initialization failed.
	 * @throws Exception
	 */
	public Object createDAOFromClassName(String className) throws Exception {
		if (className == null) {
			return null;
		}
		Class classDefinition = null;
		try {
			classDefinition = Class.forName(className + getDatabaseType());
			if (log.isDebugEnabled())
				log.debug("Found implementing class: " + className + getDatabaseType());
		} catch (ClassNotFoundException cnfe) {
			try {
				classDefinition = Class.forName(className + "Basic");
				if (log.isDebugEnabled())
					log.debug("Found implementing class: " + className + "Basic");
			} catch (ClassNotFoundException cnfe2) {
				classDefinition = Class.forName(className);
				if (log.isDebugEnabled())
					log.debug("Found implementing class: " + className);
			}
		}
		Object object = classDefinition.newInstance();
		DAOInterface dao = (DAOInterface) object;
		dao.initialize(getTransferMap());
		return dao;
	}

	private Properties appSettings = null;

	public Properties getAppSettings() {
		return appSettings;
	}

	public void setAppSettings(Properties appSettings) {
		this.appSettings = appSettings;
	}

	private Properties appComponents = null;

	public Properties getAppComponents() {
		return appComponents;
	}

	public void setAppComponents(Properties appComponents) {
		this.appComponents = appComponents;
	}

	private Properties appComponentsDefaults = null;

	public Properties getAppComponentsDefaults() {
		return appComponentsDefaults;
	}

	public void setAppComponentsDefaults(Properties appComponentsDefaults) {
		this.appComponentsDefaults = appComponentsDefaults;
	}

	private Properties hibernateSettings = null;

	public Properties getHibernateSettings() {
		return hibernateSettings;
	}

	public void setHibernateSettings(Properties hibernateSettings) {
		this.hibernateSettings = hibernateSettings;
	}

	private SessionFactory sessionFactory = null;

	public SessionFactory getSessionFactory() {
		return sessionFactory;
	}

	public void setSessionFactory(SessionFactory sessionFactory) {
		this.sessionFactory = sessionFactory;
	}

	public Session startSession() {
		return null;
	}

	public Session getSession() {
		return null;
	}

	public void closeSession() {
	}

	public void beginTransaction() {
	}

	public void commitTransaction() throws HibernateException {
	}
	
	public void rollbackTransactionKeepingSessionOpen() {
	}

	public void rollbackTransaction() {
	}

	public void evictCache() {
	}

}
